"""GuacaMol baselines training utilities."""

from dataclasses import dataclass, field
from typing import Any, Dict

from ..core import TrainingPipeline, TrainingPipelineArguments


class GuacaMolBaselinesTrainingPipeline(TrainingPipeline):
    """GuacaMol Baselines training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for GuacaMol Baselines training.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Raises:
            NotImplementedError: the generic trainer does not implement the pipeline.
        """
        raise NotImplementedError


@dataclass
class GuacaMolDataArguments(TrainingPipelineArguments):
    """Arguments related to data loading."""

    __name__ = "dataset_args"

    train_smiles_filepath: str = field(
        metadata={"help": "Path of SMILES file for Training."}
    )
    test_smiles_filepath: str = field(
        metadata={"help": "Path of SMILES file for Validation."}
    )


@dataclass
class GuacaMolSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to GuacaMol trainer."""

    __name__ = "saving_args"

    model_filepath: str = field(metadata={"help": "Path to the model file."})
    model_config_filepath: str = field(
        metadata={"help": "Path to the model config file."}
    )



"""Argument parser unit tests."""

from dataclasses import dataclass, field
from typing import Union

from gt4sd.cli.argument_parser import ArgumentParser


@dataclass
class TestArguments:

    int_arg: int = field(default=0)

    float_arg: float = field(default=0.0)

    str_arg: str = field(default="test")

    bool_arg: bool = field(default=True)

    int_none_arg: Union[int, None] = field(default=None)

    float_none_arg: Union[float, None] = field(default=None)

    str_none_arg: Union[str, None] = field(default=None)

    bool_none_arg: Union[bool, None] = field(default=None)


def test_int_default():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert isinstance(args[0].int_arg, int)
    assert args[0].int_arg == 0


def test_float_default():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert isinstance(args[0].float_arg, float)
    assert args[0].float_arg == 0.0


def test_str_default():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert isinstance(args[0].str_arg, str)
    assert args[0].str_arg == "test"


def test_bool_default():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert isinstance(args[0].bool_arg, bool)
    assert args[0].bool_arg is True


def test_int_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--int_arg", "1"])

    assert isinstance(args[0].int_arg, int)
    assert args[0].int_arg == 1


def test_float_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--float_arg", "1.0"])

    assert isinstance(args[0].float_arg, float)
    assert args[0].float_arg == 1.0


def test_str_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--str_arg", "my_test"])

    assert isinstance(args[0].str_arg, str)
    assert args[0].str_arg == "my_test"


def test_bool_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--bool_arg", "False"])

    assert isinstance(args[0].bool_arg, bool)
    assert args[0].bool_arg is False


def test_bool_int_assigned():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--bool_arg", "0"])

    assert isinstance(args[0].bool_arg, bool)
    assert args[0].bool_arg is False


def test_int_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert args[0].int_none_arg is None


def test_float_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert args[0].float_none_arg is None


def test_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert args[0].str_none_arg is None


def test_bool_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses([])

    assert args[0].bool_none_arg is None


def test_int_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--int_none_arg", ""])

    assert args[0].int_none_arg is None


def test_float_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--float_none_arg", ""])

    assert args[0].float_none_arg is None


def test_str_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--str_none_arg", ""])

    assert args[0].str_none_arg is None


def test_bool_str_none():

    parser = ArgumentParser((TestArguments))  # type: ignore

    args = parser.parse_args_into_dataclasses(["--bool_none_arg", ""])

    assert args[0].bool_none_arg is None



"""Parsing utilties."""

import argparse
import ast
from typing import Union


def str2bool(s: Union[str, bool]) -> bool:
    """Convert a string into a bool.

    Args:
        s: a string representation of a boolean.

    Raises:
        argparse.ArgumentTypeError: in case the conversion is failing.

    Returns:
        the converted value.
    """
    if isinstance(s, bool):
        return s
    if s.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif s.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def convert_string_to_class(s: str):
    """Convert a string into a python object.

    Fallback to ast in case of unexpected strings.

    Args:
        s: a string.

    Returns:
        the converted python object.
    """
    if s.lower() == "true":
        return True
    elif s.lower() == "false":
        return False
    elif s.lower() == "none":
        return None
    elif s:
        try:
            return ast.literal_eval(s)
        except (ValueError, SyntaxError):
            return s



"""Implementation of the zero-shot classifier."""

import json
import logging
import os
from typing import List, Optional, Union

import torch
from transformers import pipeline

from ....frameworks.torch import device_claim

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ZeroShotClassifier:
    """
    Zero-shot classifier based on the HuggingFace pipeline leveraging MLNI.
    """

    def __init__(
        self,
        resources_path: str,
        model_name: str,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """Initialize ZeroShotClassifier.

        Args:
            resources_path: path where to load hypothesis, candidate labels and, optionally, the model.
            model_name: name of the model to load from the cache or download from HuggingFace.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        device = device_claim(device)
        self.device = -1 if device.type == "cpu" else int(device.type.split(":")[1])
        self.resources_path = resources_path
        self.model_name = model_name
        self.load_pipeline()

    def load_pipeline(self) -> None:
        """Load zero shot classification MLNI pipeline."""
        metadata_filepath = os.path.join(self.resources_path, "metadata.json")
        if os.path.exists(metadata_filepath):
            with open(metadata_filepath) as fp:
                metadata = json.load(fp)
                self.labels = metadata["labels"]
                self.hypothesis_template = metadata["hypothesis_template"]
            self.model_name_or_path = os.path.join(self.resources_path, self.model_name)
            if not os.path.exists(self.model_name_or_path):
                logger.info(
                    f"no model named {self.model_name_or_path} in cache, using HuggingFace"
                )
                self.model_name_or_path = self.model_name
        else:
            message = f"could not retrieve the MLNI pipeline from the cache: {metadata_filepath} does not exists!"
            logger.error(message)
            raise ValueError(message)
        self.model = pipeline(
            "zero-shot-classification",
            model=self.model_name_or_path,
            device=self.device,
        )

    def predict(self, text: str) -> List[str]:
        """Get sorted classification labels.

        Args:
            text: text to classify.

        Returns:
            labels sorted by score from highest to lowest.
        """
        return self.model(
            text,
            candidate_labels=self.labels,
            hypothesis_template=self.hypothesis_template,
        )["labels"]



"""Moses baselines training utilities."""

from dataclasses import dataclass, field
from typing import Any, Dict

from ..core import TrainingPipeline, TrainingPipelineArguments


class MosesTrainingPipeline(TrainingPipeline):
    """PyTorch lightining training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        common_args: Dict[str, Any],
    ) -> None:
        """Generic training function for GuacaMol Baselines training.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            common_args: common arguments passed to the configuration.

        Raises:
            NotImplementedError: the generic trainer does not implement the pipeline.
        """
        raise NotImplementedError


@dataclass
class MosesDataArguments(TrainingPipelineArguments):
    """Arguments related to Moses data loading."""

    __name__ = "dataset_args"
    train_load: str = field(
        metadata={"help": "Input data in csv format used for training."}
    )
    val_load: str = field(
        metadata={"help": "Input data in csv format used for validation."}
    )


@dataclass
class MosesTrainingArguments(TrainingPipelineArguments):
    """Arguments related to Moses trainer."""

    __name__ = "training_args"
    model_save: str = field(metadata={"help": "Path where the trained model is saved."})
    log_file: str = field(metadata={"help": "Path where to save the the logs."})
    config_save: str = field(metadata={"help": "Path for the config."})
    vocab_save: str = field(metadata={"help": "Path to save the model vocabulary."})
    save_frequency: int = field(
        default=1, metadata={"help": "How often to save the model."}
    )
    seed: int = field(
        default=0, metadata={"help": "Seed used for random number generation."}
    )
    device: str = field(
        default="cpu",
        metadata={"help": "Device to run: 'cpu' or 'cuda:<device number>'"},
    )


@dataclass
class MosesSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to PaccMann trainer."""

    __name__ = "saving_args"

    model_path: str = field(metadata={"help": "Path where the model is stored."})
    config_path: str = field(metadata={"help": "Path where the config is stored."})
    vocab_path: str = field(metadata={"help": "Path where the vocab is stored."})



"""Utilities used in the tests."""

from functools import lru_cache

from pydantic import BaseSettings


class GT4SDTestSettings(BaseSettings):
    """Utility variables for the tests setup."""

    gt4sd_s3_host: str = "s3.mil01.cloud-object-storage.appdomain.cloud"
    gt4sd_s3_access_key: str = "a19f93a1c67949f1a31db38e58bcb7e8"
    gt4sd_s3_secret_key: str = "5748375c761a4f09c30a68cd15e218e3b27ca3e2aebd7726"
    gt4sd_s3_secure: bool = True
    gt4sd_ci: bool = False

    class Config:
        # immutable and in turn hashable, that is required for lru_cache
        frozen = True

    @staticmethod
    @lru_cache(maxsize=None)
    def get_instance() -> "GT4SDTestSettings":
        return GT4SDTestSettings()



"""Controlled sampling of concatenated encodings via Gaussian Process."""

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

import numpy as np
import pandas as pd
import torch
from skopt import gp_minimize
from skopt.space import Real

from .....frameworks.torch import device_claim

Point = Union[np.ndarray, torch.Tensor, pd.Series]
MINIMUM_REAL = np.finfo(dtype=np.float32).min
MAXIMUM_REAL = np.finfo(dtype=np.float32).max


def point_to_tensor(point: Point) -> torch.Tensor:
    """Convert point to tensor.

    Args:
        point: a point.

    Returns:
        tensor representing a point.
    """
    return (
        point.clone().detach().float()
        if isinstance(point, torch.Tensor)
        else torch.tensor(point).float()
    )


@dataclass
class Representation:
    """
    A generic representation for a composition problem.

    Attributes:
        model: a torch module for decoding.
        z_dimension: dimension of the latent space.
        fixed_representation: fixed representation in the latent space.
        z_index: slice for indexing a point to represent the latent space.
    """

    model: torch.nn.Module
    z_dimension: int
    fixed_representation: Optional[torch.Tensor] = None
    z_index: Optional[slice] = None

    def decode(self, z: Point) -> Any:
        """Decode the representation from the latent space.

        Args:
            z: a point in the latent space.

        Returns:
            the decoded representation.
        """
        z = torch.unsqueeze(point_to_tensor(z), dim=0)
        reconstructed = self.model.decode(z)  # type: ignore
        return reconstructed

    def deserialize(
        self, filepath: str, device: Optional[Union[torch.device, str]] = None
    ) -> None:
        """
        Deserialize a representation from file.

        Args:
            filepath: path to the serialized represantation.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.

        Returns:
            the representation object.
        """
        device = device_claim(device)
        weights = torch.load(filepath, map_location=device)
        self.model = self.model.to(device)  # type:ignore
        self.model.load_state_dict(weights)


class PropertyPredictor:
    """Property prediction class.

    Attributes:
        input_representations: order of the input representations.
    """

    input_representations: Optional[List[str]] = None

    def __call__(self, z: Point) -> float:
        """Call the property predictor on the point.

        Args:
            z: the point.

        Returns:
            the predicted property.
        """
        raise NotImplementedError("Propery prediction not implemented")


class Scaler:
    """Scaler class."""

    def __call__(self, example: Any) -> Any:
        """Scale the example appropriately.

        Args:
            example: an example prior to scaling.

        Returns:
            the scaled example.
        """
        raise NotImplementedError("Scaler not implemented not implemented")


RepresentationsDict = Dict[str, Representation]


class Objective:
    """
    Objective function for representations.
    """

    def __init__(
        self,
        targets: Dict[str, float],
        property_predictors: Dict[str, PropertyPredictor],
        representations: RepresentationsDict,
        representation_order: List[str] = None,
        scalers: Optional[Dict[str, Scaler]] = None,
        weights: Optional[Dict[str, float]] = None,
        custom_score_function: Optional[
            Callable[[Point, RepresentationsDict, Optional[Dict[str, Scaler]]], float]
        ] = None,
        custom_score_weight: float = 1.0,
        minimize: bool = True,
    ):
        """Constructs an objective function.

        Args:
            targets: a dictionary of target values.
            property_predictors: a dictionary of target property predictors.
            representations: a dictionary of decodeable representations.
            representation_order: order of the representations. Defaults to None, a.k.a., lexicographic order.
            scalers: scalers for represantation features. Defaults to None, a.k.a., no scaling.
            weights: weights for each the target. Defaults to None, a.k.a., targets evenly weigthed.
            custom_score_function: a custom score function to apply on decoded representations. Defaults to None, a.k.a., no custom score.
            custom_score_weight: weight for the custom score. Defaults to 1.0.
            minimize: whether the objective needs to be minimized. Defaults to True.
        """
        self.targets = targets
        self.property_predictors = property_predictors
        self.representations = representations
        if representation_order is None:
            self.representation_order = sorted(list(representations.keys()))
        else:
            self.representation_order = representation_order
        self.scalers = scalers
        self.weights = weights
        self.custom_score_function = custom_score_function
        self.custom_score_weight = custom_score_weight
        self.minimize = minimize

        if self.weights is None:
            weights_dictionary = dict()
            for target in self.targets.keys():
                weights_dictionary[target] = 1.0
            self.weights = weights_dictionary

    def construct_property_representation(
        self, z: torch.Tensor, property_name: str
    ) -> torch.Tensor:
        """Construct a representation for a specific property.

        The encoded point and fixed encodings (or slices thereof) if available
        are concatenated in the right order.

        Todo:
            Check explanation and improve it.

        Args:
            z: the point.
            property_name: name of the property for which to construct the representation.

        Returns:
            representation for which a specific property can be predicted.
        """
        # TODO make generic for self.representations: RepresentationsDict
        # defer this to the configuration, or some other place that defines how representations belong together
        propery_predictor = self.property_predictors[property_name]
        if propery_predictor.input_representations:
            representation_names = propery_predictor.input_representations
        else:
            representation_names = self.representation_order
        z_list = []
        for representation_name in representation_names:
            representation = self.representations[representation_name]
            if representation.fixed_representation:
                z_list.append(representation.fixed_representation)
            else:
                z_list.append(z[representation.z_index])
        z_latent = torch.cat(z_list)
        return z_latent.reshape(-1, z_latent.shape[0])

    def evaluate(self, z: Point) -> float:
        """Evaluate objective function for a point in latent space.

        Args:
            z: the point.

        Returns:
            the score of the point.
        """
        z = point_to_tensor(z)
        # predict all properties
        predicted_properties = dict()
        with torch.no_grad():
            for property_name in self.targets.keys():
                property_predictor = self.property_predictors[property_name]
                latent_z = self.construct_property_representation(
                    z=z, property_name=property_name
                )
                predicted_properties[property_name] = property_predictor(latent_z)

                # TODO aggregate the following scores over different properties,
                # or really only keep last one?
                if self.custom_score_function:
                    custom_score = self.custom_score_function(
                        latent_z,
                        self.representations,
                        self.scalers,
                    )
                else:
                    custom_score = 0

        scores = []
        for property_name, predicted_property in predicted_properties.items():
            scores.append(abs(self.targets[property_name] - predicted_property))
        score = sum(scores)

        # this is to penalize `custom_score` the non normalization
        total_score = score + self.custom_score_weight * custom_score

        if not self.minimize:
            score = -1 * total_score
        return total_score


class GaussianProcessRepresentationsSampler:
    def __init__(
        self,
        targets: Dict[str, float],
        property_predictors: Dict[str, PropertyPredictor],
        representations: RepresentationsDict,
        representation_order: Optional[List[str]] = None,
        bounds: Optional[
            Dict[str, Union[List[Tuple[float, float]], Tuple[float, float]]]
        ] = None,
        # TODO Any should be type of scaler; default to lambda returning 0?
        scalers: Optional[Dict[str, Scaler]] = None,
        weights: Optional[Dict[str, float]] = None,
        custom_score_function: Optional[
            Callable[[Point, RepresentationsDict, Optional[Dict[str, Scaler]]], float]
        ] = None,
        custom_score_weight: float = 1.0,
        minimize: bool = True,
        random_state: int = 42,
        random_starts: int = 10,
    ):
        """Constucts a GaussianProcessRepresentationsSampler.

        Args:
            targets: a dictionary of target values.
            property_predictors: a dictionary of target property predictors.
            representations: a dictionary of decodeable representations.
            representation_order: order of the representations. Defaults to None, a.k.a., lexicographic order.
            bounds: bounds for the optmization. Defaults to None, a.k.a., unbounded.
            scalers: scalers for represantation features. Defaults to None, a.k.a., no scaling.
            weights: weights for each the target. Defaults to None, a.k.a., targets evenly weigthed.
            custom_score_function: a custom score function to apply on decoded representations. Defaults to None, a.k.a., no custom score.
            custom_score_weight: weight for the custom score. Defaults to 1.0.
            minimize: whether the objective needs to be minimized. Defaults to True.
            random_state: random state. Defaults to 42.
            random_starts: number of random restarts. Defaults to 10.
        """
        self.targets = targets
        self.property_predictors = property_predictors
        self.representations = representations
        self.representation_order = representation_order
        if self.representation_order is None:
            self.representation_order = sorted(list(representations.keys()))
        self.scalers = scalers
        self.weigths = weights
        self.custom_score_function = custom_score_function
        self.custom_score_weight = custom_score_weight
        self.minimize = minimize
        self.random_state = random_state
        self.random_starts = random_starts
        self.set_bounds(bounds)
        self.dimensions = self.define_dimensions(self.representation_order)

    def _get_bounds(
        self, minimum_value: float, maximum_value: float, z_dimension: int
    ) -> List[Tuple[float, float]]:
        """
        Define a list of bounds for an hypercube.

        Args:
            minimum_value: minimum value.
            maximum_value: maximum value.
            z_dimension: dimension of the hypercube.

        Returns:
            the list of bounds.
        """
        return [(minimum_value, maximum_value) for _ in range(z_dimension)]

    def set_bounds(
        self,
        bounds: Optional[
            Dict[str, Union[List[Tuple[float, float]], Tuple[float, float]]]
        ] = None,
    ) -> None:
        """Set the bounds for the optimization.

        Args:
            bounds: bounds for the optmization. Defaults to None, a.k.a., unbounded.
        """
        self.bounds = bounds if bounds else dict()
        for representation_name, bounds in self.bounds.items():  # type:ignore
            z_dimension = self.representations[representation_name].z_dimension
            if isinstance(bounds, tuple):
                self.bounds[representation_name] = self._get_bounds(
                    bounds[0], bounds[1], z_dimension
                )
            else:
                self.bounds[representation_name] = bounds  # type:ignore
        for representation_name in self.representations.keys() - self.bounds.keys():
            z_dimension = self.representations[representation_name].z_dimension
            self.bounds[representation_name] = self._get_bounds(
                MINIMUM_REAL, MAXIMUM_REAL, z_dimension
            )

    def define_dimensions(self, representation_order: List[str]) -> List[Real]:
        """Define the dimensions of the optimization space.

        Args:
            representation_order: order of the representations.

        Returns:
            a list of dimensions.
        """
        dimensions = []
        latent_index = 0
        for representation_name in representation_order:
            representation = self.representations[representation_name]
            representation_bounds = self.bounds[representation_name]
            representation.z_index = slice(
                latent_index, latent_index + representation.z_dimension
            )
            latent_index += representation.z_dimension
            dimensions.extend(
                [
                    Real(lower_bound, upper_bound)  # type:ignore
                    for lower_bound, upper_bound in representation_bounds
                ]
            )
        return dimensions

    def optimize(
        self,
        targets: Optional[Dict[str, float]] = None,
        relevant_representations: Optional[List[str]] = None,
        representation_order: Optional[List[str]] = None,
        z0: Optional[Point] = None,
        number_of_points: int = 1,
        number_of_steps: int = 50,
        random_state: int = 42,
        verbose: bool = False,
        weights: Optional[Dict[str, float]] = None,
        acquisition_method: str = "PI",
    ) -> List[Dict[str, Any]]:
        """Run the optimization.

        Args:
            targets: a dictionary of target values. Defaults to None, a.k.a., use the ones defined at construction time.
            relevant_representations: list of relevant representations to be optimized. Defaults to None, a.k.a., inferred from non fixed representations.
            representation_order: order of the representations. Defaults to None, a.k.a., use the one defined at construction time.
            z0: the starting point for the optimization. Defaults to None, a.k.a., perform random starts.
            number_of_points: number of optimal points to return. Defaults to 1.
            number_of_steps: number of optimization steps. Defaults to 50.
            random_state: random state. Defaults to 42.
            verbose: control verbosity. Defaults to False.
            weights: weights for each the target. Defaults to None, a.k.a., use the ones defined at construction time.
            acquisition_method: acquisition method to use in the Gaussian Process optimization. Defaults to "PI". More details at: https://scikit-optimize.github.io/stable/modules/generated/skopt.gp_minimize.html.

        Raises:
            NotImplementedError: invalid acquisition function.

        Returns:
            list of orderd optimal points with decoded relevant representations.
        """
        if representation_order is None:
            representation_order = self.representation_order
        dimensions = self.define_dimensions(cast(List[str], representation_order))

        np.random.seed(random_state)

        if acquisition_method not in ["PI", "LCB", "EI", "gp_hedge", "EIps", "PIps"]:
            raise NotImplementedError("Give valid acquisition function")

        if targets is None:
            targets = self.targets

        if weights is None:
            weights = self.weigths

        objective = Objective(
            targets=targets,
            property_predictors=self.property_predictors,
            representations=self.representations,
            representation_order=representation_order,
            scalers=self.scalers,
            weights=weights,
            custom_score_function=self.custom_score_function,
            custom_score_weight=self.custom_score_weight,
            minimize=self.minimize,
        )

        y0 = None
        random_starts: Optional[int]
        if z0 is None:
            random_starts = self.random_starts
        else:
            random_starts = None

        gaussian_process_results = gp_minimize(
            objective.evaluate,
            dimensions,
            n_calls=number_of_steps,
            n_random_starts=random_starts,
            x0=z0,
            y0=y0,
            acq_func=acquisition_method,
            random_state=np.random.RandomState(random_state),
            verbose=verbose,
        )

        objective_values, points = zip(
            *sorted(
                zip(
                    gaussian_process_results.func_vals, gaussian_process_results.x_iters
                )
            )
        )

        results_list = []

        if relevant_representations is None:
            relevant_representations = [
                representation_name
                for representation_name, representation in self.representations.items()
                if representation.fixed_representation is None
            ]

        for point, objective_value in zip(
            points[:number_of_points], objective_values[:number_of_points]
        ):
            optimization_result = {"objective": objective_value, "z": point}
            for representation_name in relevant_representations:
                representation = self.representations[representation_name]
                z = point[representation.z_index]
                optimization_result[representation_name] = representation.decode(z)

            results_list.append(optimization_result)

        return results_list


class Generator:
    def generate_samples(self, target: Any) -> List[Any]:
        """Generate samples.

        Args:
            target: target for generation.

        Returns:
            samples generated.
        """
        raise NotImplementedError("Generate samples not implemented.")



"""PaccMann VAE trainer unit tests."""

import shutil
import tempfile
from typing import Any, Dict, cast

import pkg_resources

from gt4sd.training_pipelines import (
    TRAINING_PIPELINE_MAPPING,
    PaccMannVAETrainingPipeline,
)

template_config = {
    "model_args": {
        "n_layers": 1,
        "bidirectional": False,
        "rnn_cell_size": 64,
        "latent_dim": 32,
        "stack_width": 8,
        "stack_depth": 8,
        "decoder_search": "sampling",
        "dropout": 0.2,
        "generate_len": 50,
        "kl_growth": 0.003,
        "input_keep": 0.85,
        "test_input_keep": 1.0,
        "temperature": 0.8,
        "embedding": "one_hot",
        "vocab_size": 380,
        "pad_index": 0,
        "embedding_size": 380,
    },
    "dataset_args": {
        "add_start_stop_token": True,
        "selfies": True,
        "num_workers": 1,
        "pin_memory": False,
    },
    "training_args": {
        "epochs": 1,
        "batch_size": 4,
        "learning_rate": 0.0005,
        "optimizer": "adam",
        "log_interval": 2,
        "save_interval": 2,
        "eval_interval": 2,
        "model_path": "/tmp/paccmann_vae",
        "training_name": "paccmann-vae-test",
    },
}


def test_train():

    pipeline = TRAINING_PIPELINE_MAPPING.get("paccmann-vae-trainer")

    assert pipeline is not None

    TEMPORARY_DIRECTORY = tempfile.mkdtemp()

    test_pipeline = cast(PaccMannVAETrainingPipeline, pipeline())

    config: Dict[str, Any] = template_config.copy()

    config["training_args"]["model_path"] = TEMPORARY_DIRECTORY

    file_path = pkg_resources.resource_filename(
        "gt4sd",
        "training_pipelines/tests/molecules.smi",
    )

    config["dataset_args"]["train_smiles_filepath"] = file_path
    config["dataset_args"]["test_smiles_filepath"] = file_path
    test_pipeline.train(**config)

    shutil.rmtree(TEMPORARY_DIRECTORY)



"""PaccMannRL initialization."""

from .core import (
    PaccMannRL,
    PaccMannRLOmicBasedGenerator,
    PaccMannRLProteinBasedGenerator,
)

__all__ = [
    "PaccMannRL",
    "PaccMannRLProteinBasedGenerator",
    "PaccMannRLOmicBasedGenerator",
]



"""PaccMannRL tests."""

import pickle
from typing import ClassVar, Type

import numpy as np
import pytest

from gt4sd.algorithms.conditional_generation.paccmann_rl import (
    PaccMannRL,
    PaccMannRLOmicBasedGenerator,
    PaccMannRLProteinBasedGenerator,
)
from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.registry import ApplicationsRegistry


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            PaccMannRLProteinBasedGenerator,
            "conditional_generation",
            "materials",
            PaccMannRL.__name__,
        ),
        (
            PaccMannRLOmicBasedGenerator,
            "conditional_generation",
            "materials",
            PaccMannRL.__name__,
        ),
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [
        (PaccMannRLProteinBasedGenerator),
        (PaccMannRLOmicBasedGenerator),
    ],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [
        (PaccMannRLProteinBasedGenerator),
        (PaccMannRLOmicBasedGenerator),
    ],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert "v0" in versions


@pytest.mark.parametrize(
    "config, example_target, algorithm",
    [
        (
            PaccMannRLProteinBasedGenerator,
            "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
            PaccMannRL,
        ),
        (
            PaccMannRLOmicBasedGenerator,
            np.random.rand(2128),
            PaccMannRL,
        ),
        (
            PaccMannRLOmicBasedGenerator,
            f"[{','.join(map(str, np.random.rand(2128)))}]",
            PaccMannRL,
        ),
    ],
)
def test_generation_via_import(config, example_target, algorithm):
    paccmann_rl = algorithm(configuration=config(), target=example_target)
    items = list(paccmann_rl.sample(5))
    assert len(items) == 5


@pytest.mark.parametrize(
    "algorithm_application, target",
    [
        (
            PaccMannRLProteinBasedGenerator.__name__,
            "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
        ),
    ],
)
def test_generation_via_registry(target, algorithm_application):
    paccmann_rl = ApplicationsRegistry.get_application_instance(
        target=target,
        algorithm_type="conditional_generation",
        domain="materials",
        algorithm_name=PaccMannRL.__name__,
        algorithm_application=algorithm_application,
        generated_length=5,
    )
    items = list(paccmann_rl.sample(5))
    assert len(items) == 5


@pytest.mark.parametrize(
    "config_class",
    [
        (PaccMannRLProteinBasedGenerator),
        (PaccMannRLOmicBasedGenerator),
    ],
)
def test_configuration_pickable(config_class: Type[AlgorithmConfiguration]):
    # implementation
    obj = config_class(algorithm_version="test")

    # ---
    import inspect

    inspect.getmodule(config_class)
    # ---
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)
    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj

    # registered
    Config = ApplicationsRegistry.get_application(
        algorithm_type="conditional_generation",
        domain="materials",
        algorithm_name=PaccMannRL.__name__,
        algorithm_application=config_class.__name__,
    ).configuration_class

    obj = Config(algorithm_version="test")
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)

    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj



"""Make pytest fixtures available to multiple test directories."""

import pytest


@pytest.fixture
def mock_wrong_s3_env(monkeypatch):
    """Changes an environment variable to break the s3 connection."""
    monkeypatch.setenv("GT4SD_S3_SECRET_KEY", "(╯°□°）╯︵ ┻━┻")



"""Implementation of MolGX conditional generators."""

import logging
import os
from typing import Any, Dict, List

from ....extras import EXTRAS_ENABLED

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if EXTRAS_ENABLED:
    from AMD_Analytics.amdsdk import AMDsdk

    class MolGXGenerator:
        """Interface for MolGX generator."""

        def __init__(
            self,
            resources_path: str,
            tag_name: str,
            homo_energy_value: float = -0.25,
            lumo_energy_value: float = 0.08,
            use_linear_model: bool = True,
            number_of_candidates: int = 2,
            maximum_number_of_candidates: int = 3,
            maximum_number_of_solutions: int = 3,
            maximum_number_of_nodes: int = 50000,
            beam_size: int = 2000,
            without_estimate: bool = True,
            use_specific_rings: bool = True,
            use_fragment_const: bool = False,
        ) -> None:
            """Instantiate a MolGX generator.

            Args:
                resources_path: path to the resources for model loading.
                tag_name: tag for the pretrained model.
                homo_energy_value: target HOMO energy value. Defaults to -0.25.
                lumo_energy_value: target LUMO energy value. Defaults to 0.08.
                use_linear_model: linear model usage. Defaults to True.
                number_of_candidates: number of candidates to consider. Defaults to 2.
                maximum_number_of_candidates: maximum number of candidates to consider. Defaults to 3.
                maximum_number_of_solutions: maximum number of solutions. Defaults to 3.
                maximum_number_of_nodes: maximum number of nodes in the graph exploration. Defaults to 50000.
                beam_size: size of the beam during search. Defaults to 2000.
                without_estimate: disable estimates. Defaults to True.
                use_specific_rings: flag to indicate whether specific rings are used. Defaults to True.
                use_fragment_const: using constant fragments. Defaults to False.

            Raises:
                RuntimeError: in the case extras are disabled.
            """
            if not EXTRAS_ENABLED:
                raise RuntimeError("Can't instantiate MolGXGenerator, extras disabled!")

            # loading artifacts
            self.resources_path = resources_path
            self.tag_name = tag_name
            self.amd = self.load_molgx(self.resources_path, self.tag_name)
            self.molecules_data, self.target_property = self.amd.LoadPickle("model")
            # algorithm parameters
            self._homo_energy_value = homo_energy_value
            self._lumo_energy_value = lumo_energy_value
            self._use_linear_model = use_linear_model
            self._number_of_candidates = number_of_candidates
            self._maximum_number_of_candidates = maximum_number_of_candidates
            self._maximum_number_of_solutions = maximum_number_of_solutions
            self._maximum_number_of_nodes = maximum_number_of_nodes
            self._beam_size = beam_size
            self._without_estimate = without_estimate
            self._use_specific_rings = use_specific_rings
            self._use_fragment_const = use_fragment_const
            self._parameters = self._create_parameters_dictionary()

        @staticmethod
        def load_molgx(resource_path: str, tag_name: str) -> AMDsdk:
            """Load MolGX model.

            Args:
                resource_path: path to the resources for model loading.
                tag_name: tag for the pretrained model.

            Returns:
                MolGX model SDK.
            """
            return AMDsdk(
                dir_pickle=os.path.join(resource_path, "pickle"),
                dir_data=os.path.join(resource_path, "data"),
                tag_data=tag_name,
            )

        def _create_parameters_dictionary(self) -> Dict[str, Any]:
            """Create parameters dictionary.

            Returns:
                the parameters to run MolGX.
            """
            self.target_property["homo"] = (self.homo_energy_value,) * 2
            self.target_property["lumo"] = (self.lumo_energy_value,) * 2
            parameters: Dict[str, Any] = {}
            parameters["target_property"] = self.target_property
            parameters["use_linear_model"] = self.use_linear_model
            parameters["num_candidate"] = self.number_of_candidates
            parameters["max_candidate"] = self.maximum_number_of_candidates
            parameters["max_solution"] = self.maximum_number_of_solutions
            parameters["max_node"] = self.maximum_number_of_nodes
            parameters["beam_size"] = self.beam_size
            parameters["without_estimate"] = self.without_estimate
            parameters["use_specific_rings"] = self.use_specific_rings
            parameters["use_fragment_const"] = self.use_fragment_const
            return parameters

        @property
        def homo_energy_value(self) -> float:
            return self._homo_energy_value

        @homo_energy_value.setter
        def homo_energy_value(self, value: float) -> None:
            self._homo_energy_value = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def lumo_energy_value(self) -> float:
            return self._lumo_energy_value

        @lumo_energy_value.setter
        def lumo_energy_value(self, value: float) -> None:
            self._lumo_energy_value = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def use_linear_model(self) -> bool:
            return self._use_linear_model

        @use_linear_model.setter
        def use_linear_model(self, value: bool) -> None:
            self._use_linear_model = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def number_of_candidates(self) -> int:
            return self._number_of_candidates

        @number_of_candidates.setter
        def number_of_candidates(self, value: int) -> None:
            self._number_of_candidates = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def maximum_number_of_candidates(self) -> int:
            return self._maximum_number_of_candidates

        @maximum_number_of_candidates.setter
        def maximum_number_of_candidates(self, value: int) -> None:
            self._maximum_number_of_candidates = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def maximum_number_of_solutions(self) -> int:
            return self._maximum_number_of_solutions

        @maximum_number_of_solutions.setter
        def maximum_number_of_solutions(self, value: int) -> None:
            self._maximum_number_of_solutions = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def maximum_number_of_nodes(self) -> int:
            return self._maximum_number_of_nodes

        @maximum_number_of_nodes.setter
        def maximum_number_of_nodes(self, value: int) -> None:
            self._maximum_number_of_nodes = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def beam_size(self) -> int:
            return self._beam_size

        @beam_size.setter
        def beam_size(self, value: int) -> None:
            self._beam_size = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def without_estimate(self) -> bool:
            return self._without_estimate

        @without_estimate.setter
        def without_estimate(self, value: bool) -> None:
            self._without_estimate = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def use_specific_rings(self) -> bool:
            return self._use_specific_rings

        @use_specific_rings.setter
        def use_specific_rings(self, value: bool) -> None:
            self._use_specific_rings = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def use_fragment_const(self) -> bool:
            return self._use_fragment_const

        @use_fragment_const.setter
        def use_fragment_const(self, value: bool) -> None:
            self._use_fragment_const = value
            self.parameters = self._create_parameters_dictionary()

        @property
        def parameters(self) -> Dict[str, Any]:
            return self._parameters

        @parameters.setter
        def parameters(self, value: Dict[str, Any]) -> None:
            parameters = self._create_parameters_dictionary()
            parameters.update(value)
            self._parameters = parameters

        def generate(self) -> List[str]:
            """Sample random molecules.

            Returns:
                sampled molecule (SMILES).
            """
            # generate molecules
            logger.info(
                f"running MolGX with the following parameters: {self.parameters}"
            )
            molecules_df = self.amd.GenMols(self.molecules_data, self.parameters)
            logger.info("MolGX run completed")
            return molecules_df["SMILES"].tolist()

else:
    logger.warning("install AMD_analytcs extras to use MolGX")



"""Graph GA implementation."""

from guacamol_baselines.graph_ga.goal_directed_generation import GB_GA_Generator


class GraphGA:
    def __init__(
        self,
        smi_file,
        mutation_rate: float,
        population_size: int,
        offspring_size: int,
        n_jobs: int,
        random_start: bool,
        generations: int,
        patience: int,
    ):
        """Initialize SMILESGA.

        Args:
            smi_file: path where to load hypothesis, candidate labels and, optionally, the smiles file.
            population_size: used with n_mutations for the initial generation of smiles within the population.
            n_jobs: number of concurrently running jobs.
            random_start: set to True to randomly choose list of SMILES for generating optimizied molecules.
            generations: number of evolutionary generations.
            patience: used for early stopping if population scores remains the same after generating molecules.
            mutation_rate: frequency of the new mutations in a single gene or organism over time.
            offspring_size: number of molecules to select for new population.
        """
        self.smi_file = smi_file
        self.mutation_rate = mutation_rate
        self.population_size = population_size
        self.offspring_size = offspring_size
        self.n_jobs = n_jobs
        self.random_start = random_start
        self.generations = generations
        self.patience = patience

    def get_generator(self) -> GB_GA_Generator:
        """Create an instance of the GB_GA_Generator.

        Returns:
            an instance of GB_GA_Generator.
        """
        optimiser = GB_GA_Generator(
            smi_file=self.smi_file,
            population_size=self.population_size,
            offspring_size=self.offspring_size,
            mutation_rate=self.mutation_rate,
            generations=self.generations,
            n_jobs=self.n_jobs,
            random_start=self.random_start,
            patience=self.patience,
        )
        return optimiser



#!/usr/bin/env python

"""Pytorch lightning checkpoint to HF transformers converter."""


import logging
import sys
from typing import cast

from transformers import Trainer, TrainingArguments

from ..training_pipelines.pytorch_lightning.language_modeling.core import (
    LanguageModelingSavingArguments,
)
from ..training_pipelines.pytorch_lightning.language_modeling.models import (
    LM_MODULE_FACTORY,
)
from .argument_parser import ArgumentParser, DataClassType


def convert_pl_to_hf(arguments: LanguageModelingSavingArguments) -> None:
    """Method to convert pytorch lightning checkpoint to HF transformers model.

    Args:
        arguments: a LanguageModelingSavingArguments instance that contains all the
                   needed arguments for conversion.
    Raises:
        ValueError: in case the provided training type is not supported.
    """

    if arguments.training_type is None:
        raise ValueError(
            "training_type is required for saving from pytorch lightning checkpoing."
        )

    if arguments.model_name_or_path is None:
        raise ValueError(
            "model_name_or_path is required for saving from pytorch lightning checkpoing."
        )

    if arguments.ckpt is None:
        raise ValueError(
            "ckpt is required for saving from pytorch lightning checkpoing."
        )

    training_type = arguments.training_type
    model_name_or_path = arguments.model_name_or_path
    tokenizer_name_or_path = arguments.tokenizer_name_or_path
    if tokenizer_name_or_path is None:
        tokenizer_name_or_path = model_name_or_path
    ckpt = arguments.ckpt
    output_path = arguments.hf_model_path

    if training_type not in LM_MODULE_FACTORY:
        ValueError(
            f"LM training type {training_type} is not supported. Supported types: {', '.join(LM_MODULE_FACTORY.keys())}."
        )
    model_module_class = LM_MODULE_FACTORY[training_type]

    model_module = model_module_class.load_from_checkpoint(
        ckpt,
        model_args={
            "model_name_or_path": model_name_or_path,
            "tokenizer": tokenizer_name_or_path,
        },
    )

    trainer = Trainer(
        model=model_module.model,
        tokenizer=model_module.tokenizer,
        args=TrainingArguments(output_dir=output_path),
    )
    trainer.save_model()


def main() -> None:
    """Convert pytorch lightning checkpoint to HF transformers model.

    Parsing from the command line the following parameters:
        - training type of the given checkpoint.
        - model name or path, a HF's model name.
        - tokenizer name or path, a HF's tokenizer name.
        - path of the checkpoint.
        - path where the HF model will be saved.

    Raises:
        ValueError: in case the provided training type is not supported.
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    arguments = ArgumentParser(
        cast(DataClassType, LanguageModelingSavingArguments)
    ).parse_args_into_dataclasses(return_remaining_strings=True)[0]

    convert_pl_to_hf(arguments)


if __name__ == "__main__":
    main()



"""
Implementation details for HuggingFace generation algorithms.

Parts of the implementation inspired by: https://github.com/huggingface/transformers/blob/v4.2.1/examples/text-generation/run_generation.py.
"""

import logging
import os
from typing import List, Optional, Union

import numpy as np
import torch
from transformers import (
    BasicTokenizer,
    CTRLLMHeadModel,
    CTRLTokenizer,
    GPT2LMHeadModel,
    GPT2Tokenizer,
    OpenAIGPTLMHeadModel,
    OpenAIGPTTokenizer,
    TransfoXLLMHeadModel,
    TransfoXLTokenizer,
    XLMTokenizer,
    XLMWithLMHeadModel,
    XLNetLMHeadModel,
    XLNetTokenizer,
)

from ....frameworks.torch import device_claim

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

MAXIMUM_LENGTH = int(10000)
# Padding text to help Transformer-XL and XLNet with short prompts as proposed by Aman Rusia
# in https://github.com/rusiaaman/XLNet-gen#methodology
# and https://medium.com/@amanrusia/xlnet-speaks-comparison-to-gpt-2-ea1a4e9ba39e
PREFIX = """In 1991, the remains of Russian Tsar Nicholas II and his family
(except for Alexei and Maria) are discovered.
The voice of Nicholas's young son, Tsarevich Alexei Nikolaevich, narrates the
remainder of the story. 1883 Western Siberia,
a young Grigori Rasputin is asked by his father and a group of men to perform magic.
Rasputin has a vision and denounces one of the men as a horse thief. Although his
father initially slaps him for making such an accusation, Rasputin watches as the
man is chased outside and beaten. Twenty years later, Rasputin sees a vision of
the Virgin Mary, prompting him to become a priest. Rasputin quickly becomes famous,
with people, even a bishop, begging for his blessing. <eod> </s> <eos>"""


def set_seed(seed: int = 42) -> None:
    """Set seed for all random number generators.

    Args:
        seed: seed to set. Defaults to 42.
    """
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available:
        torch.cuda.manual_seed_all(seed)  # type:ignore


def prepare_ctrl_input(tokenizer: BasicTokenizer, prompt: str, **kwargs):
    if kwargs.get("temperature", 1.0) > 0.7:
        logger.warning(
            "CTRL typically works better with lower temperatures (and lower k)."
        )

    encoded_prompt = tokenizer.encode(prompt, add_special_tokens=False)  # type:ignore
    if not any(
        encoded_prompt[0] == x for x in tokenizer.control_codes.values()  # type:ignore
    ):
        logger.warning(
            "not starting generation from a control code so you will not get good results"
        )
    return prompt


def prepare_prefix_input(tokenizer: BasicTokenizer, prompt: str, **kwargs):
    prefix = kwargs["prefix"] if kwargs.get("prefix", "") else PREFIX
    prompt = prefix + prompt
    return prompt


def adjust_length_to_model(length: int, maximum_sequence_length: int):
    """Adjust sequence length.

    Args:
        length: target length.
        maximum_sequence_length: maximum sequence length.

    Returns:
        the adjusted length.
    """
    if length < 0 and maximum_sequence_length > 0:
        logger.warning(
            f"negative length, adjusting to model supported length {maximum_sequence_length}"
        )
        length = maximum_sequence_length
    elif 0 < maximum_sequence_length < length:
        logger.warning(
            f"longer then model supported length, adjusting to {maximum_sequence_length}"
        )
        length = maximum_sequence_length
    elif length < 0:
        logger.warning(f"negative length, adjusting to maximal length {MAXIMUM_LENGTH}")
        length = MAXIMUM_LENGTH
    return length


MODEL_TYPES = {
    "gpt2": (GPT2LMHeadModel, GPT2Tokenizer, None),
    "ctrl": (CTRLLMHeadModel, CTRLTokenizer, prepare_ctrl_input),
    "openai-gpt": (OpenAIGPTLMHeadModel, OpenAIGPTTokenizer, None),
    "xlnet": (XLNetLMHeadModel, XLNetTokenizer, prepare_prefix_input),
    "transfo-xl": (TransfoXLLMHeadModel, TransfoXLTokenizer, prepare_prefix_input),
    "xlm": (XLMWithLMHeadModel, XLMTokenizer, None),
}


class Generator:
    """Implementation of a generator."""

    def __init__(
        self,
        resources_path: str,
        model_type: str,
        model_name: str,
        prompt: str,
        length: int,
        stop_token: str,
        temperature: float,
        repetition_penalty: float,
        k: int,
        p: float,
        prefix: str,
        number_of_sequences: int,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """An HuggingFace generation algorithm.

        Args:
            resources_path: path to the cache.
            model_type: type of the model.
            model_name: name of the model weights/version.
            prompt: prompt for text generation.
            length: length of the generated text.
            stop_token: stop token for text generation.
            temperature: temperature for sampling, the lower the greedier the sampling.
            repetition_penalty: primarily useful for CTRL model, where 1.2 should be used.
            k: number of top-k probability token to keep.
            p: only tokens with cumulative probabilities summing up to this value are kept.
            prefix: text defining context provided prior to the prompt.
            number_of_sequences: number of generated sequences.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        self.device = device_claim(device)
        self.resources_path = resources_path
        self.model_type = model_type
        self.model_name = model_name
        self.prompt = prompt
        self.length = length
        self.stop_token = None if stop_token == "" else stop_token
        self.temperature = temperature
        self.repetition_penalty = repetition_penalty
        self.k = k
        self.p = p
        self.prefix = prefix
        self.number_of_sequences = number_of_sequences
        self.load_model()

    def load_model(self) -> None:
        """Load a pretrained HuggingFace generation model."""
        try:
            model_class, tokenizer_class, preprocessing_function = MODEL_TYPES[
                self.model_type
            ]
        except KeyError:
            raise KeyError(f"model type: {self.model_type} not supported")
        if (
            os.path.exists(self.resources_path)
            and len(os.listdir(self.resources_path)) > 0
        ):
            model_name_or_path = self.resources_path
        else:
            model_name_or_path = self.model_name
        self.preprocessing_function = preprocessing_function
        self.tokenizer = tokenizer_class.from_pretrained(  # type:ignore
            model_name_or_path
        )
        self.model = model_class.from_pretrained(model_name_or_path)
        self.model.to(self.device)
        # adjusting length
        self.length = adjust_length_to_model(
            self.length, self.model.config.max_position_embeddings
        )

    def sample(self) -> List[str]:
        """Sample text snippets.

        Returns:
            generated text snippets.
        """
        if self.preprocessing_function is not None:
            preprocessed_prompt_text = self.preprocessing_function(
                self.tokenizer,
                self.prompt,
                prefix=self.prefix,
                temperature=self.temperature,
            )

            if self.model.__class__.__name__ in ["TransfoXLLMHeadModel"]:
                tokenizer_kwargs = {"add_space_before_punct_symbol": True}
            else:
                tokenizer_kwargs = {}

            encoded_prompt = self.tokenizer.encode(
                preprocessed_prompt_text,
                add_special_tokens=False,
                return_tensors="pt",
                **tokenizer_kwargs,
            )
        else:
            encoded_prompt = self.tokenizer.encode(
                self.prefix + self.prompt, add_special_tokens=False, return_tensors="pt"
            )

        encoded_prompt = encoded_prompt.to(self.device)

        if encoded_prompt.size()[-1] == 0:
            input_ids = None
        else:
            input_ids = encoded_prompt

        output_sequences = self.model.generate(
            input_ids=input_ids,
            max_length=self.length + len(encoded_prompt[0]),
            temperature=self.temperature,
            top_k=self.k,
            top_p=self.p,
            repetition_penalty=self.repetition_penalty,
            do_sample=True,
            num_return_sequences=self.number_of_sequences,
        )

        # NOTE: remove the batch dimension when returning multiple sequences
        if len(output_sequences.shape) > 2:
            output_sequences.squeeze_()

        generated_sequences: List[str] = []

        for generated_sequence in output_sequences:
            generated_sequence = generated_sequence.tolist()
            text = self.tokenizer.decode(
                generated_sequence, clean_up_tokenization_spaces=True
            )
            text = text[: text.find(self.stop_token) if self.stop_token else None]
            total_sequence = (
                self.prompt
                + text[
                    len(
                        self.tokenizer.decode(
                            encoded_prompt[0], clean_up_tokenization_spaces=True
                        )
                    ) :
                ]
            )
            generated_sequences.append(total_sequence)

        return generated_sequences



"""MLP predictor implementation."""

import logging
from argparse import ArgumentParser
from typing import Any, Dict, List, Optional, Tuple

from ....arg_parser.utils import str2bool
from ..base_model import GranularBaseModel
from ..loss import LOSS_FACTORY
from ..module import Mlp

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class MlpPredictor(GranularBaseModel):
    """MlpPredictor - Multi Layer Perceptron predictor."""

    def __init__(
        self,
        name: str,
        from_position: List[int],
        data: Dict[str, str],
        input_size: int,
        hidden_size: int,
        output_size: int,
        n_layers: int,
        activation: str,
        dropout: float,
        loss_function: str,
        class_weights: Optional[List[float]] = None,
        **kwargs,
    ) -> None:
        """Construct MlpPredictor.

        Args:
            name: model name.
            from_position: list of input model positions.
            data: data name mappings.
            input_size: size of the input.
            hidden_size: size of the hidden layers.
            output_size: size of the output.
            n_layers: number of layers.
            activation: name of the activation.
            dropout: dropout rate.
            loss_function: name of the loss function.
            class_weights: weights for the classes. Defaults to None, a.k.a., no weighting.

        Raises:
            ValueError: in case the provided loss function is not supported.
        """
        super().__init__(name=name, data=data)
        self.from_position = from_position
        self.target_key = name + "_" + data["target"]
        self.loss_function_name = loss_function.lower()
        if self.loss_function_name not in LOSS_FACTORY:
            raise ValueError(
                f"loss_function={self.loss_function_name} not supported. Pick a valid one: {sorted(list(LOSS_FACTORY.keys()))}"
            )
        self.loss_function = LOSS_FACTORY[self.loss_function_name]
        self.class_weights = class_weights
        self.mlp = Mlp(
            input_size=input_size,
            hidden_size=hidden_size,
            output_size=output_size,
            n_layers=n_layers,
            activation=activation,
            dropout=dropout,
        )

    def _run_step(self, x: Any, *args, **kwargs) -> Any:
        """Run a step in the model.

        Args:
            x: model input.

        Returns:
            model step output.
        """
        return self.mlp(x)

    def predict(self, x: Any, *args, **kwargs) -> Any:
        """Forward pass in the model.

        Args:
            x: model input.

        Returns:
            model output.
        """
        return self._run_step(x)

    def step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Tuple[Any, Any, Any]:
        """Training step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        output = self._run_step(input_data)
        loss = self.loss_function(output, target_data)
        logs = {f"{self.loss_function_name}_loss": loss, "loss": loss}
        return output, loss, logs

    def val_step(
        self,
        input_data: Any,
        target_data: Any,
        device: str = "cpu",
        current_epoch: int = 0,
        *args,
        **kwargs,
    ) -> Any:
        """Validation step for the model.

        Args:
            input_data: input for the step.
            target_data: target for the step.
            device: string representing the device to use. Defaults to "cpu".
            current_epoch: current epoch. Defaults to 0.

        Returns:
            a tuple containing the step output, the loss and the logs for the module.
        """
        output = self._run_step(input_data)
        loss = self.loss_function(output, target_data)
        logs = {f"{self.loss_function_name}_loss": loss, "loss": loss}

        if self.loss_function_name == "bce":
            output_label = (output > 0.5).float()
            correct_label = (output_label == target_data).float().sum()
            accuracy = correct_label / output_label.shape[0]
            logs["accuracy"] = accuracy
        return output, loss, logs

    @staticmethod
    def add_model_specific_args(
        parent_parser: ArgumentParser, name: str, *args, **kwargs
    ) -> ArgumentParser:
        """Adding to a parser model specific arguments.

        Args:
            parent_parser: patent parser.
            name: model name.

        Returns:
            updated parser.
        """
        parser = ArgumentParser(parents=[parent_parser], add_help=False)

        parser.add_argument(f"--data_path_{name}", type=str)
        parser.add_argument(f"--data_file_{name}", type=str)
        parser.add_argument(f"--dataset_type_{name}", type=str)
        parser.add_argument(f"--target_{name}", type=str)
        parser.add_argument(f"--from_position_{name}", type=int, nargs="+")
        parser.add_argument(f"--checkpoint_path_{name}", type=str)
        parser.add_argument(f"--checkpoint_model_name_{name}", type=str)
        parser.add_argument(f"--start_from_checkpoint_{name}", type=str2bool)
        parser.add_argument(f"--freeze_weights_{name}", type=str2bool)
        parser.add_argument(f"--n_layers_{name}", type=int)
        parser.add_argument(f"--activation_{name}", type=str)
        parser.add_argument(f"--dropout_{name}", type=float)
        parser.add_argument(f"--loss_function_{name}", type=str)
        parser.add_argument(f"--hidden_size_{name}", type=int)
        parser.add_argument(f"--output_size_{name}", type=int)

        return parser



"""Exceptions tests."""

from gt4sd.training_pipelines import (
    TRAINING_PIPELINE_ARGUMENTS_MAPPING,
    TRAINING_PIPELINE_NAME_METADATA_MAPPING,
    training_pipeline_name_to_metadata,
)


def test_metadata_retrieval_for_registered_pipelines_from_json():
    for name, filename in TRAINING_PIPELINE_NAME_METADATA_MAPPING.items():
        pipeline_metadata = training_pipeline_name_to_metadata(name)
        assert pipeline_metadata["training_pipeline"] == name
        assert "description" in pipeline_metadata
        assert "parameters" in pipeline_metadata
        assert "description" not in pipeline_metadata["parameters"]


def test_metadata_retrieval_for_registered_pipelines_from_dataclass():
    for name, filename in TRAINING_PIPELINE_ARGUMENTS_MAPPING.items():
        pipeline_metadata = training_pipeline_name_to_metadata(name)
        assert pipeline_metadata["training_pipeline"] == name
        assert "description" in pipeline_metadata
        assert "parameters" in pipeline_metadata
        assert "description" not in pipeline_metadata["parameters"]

        for parameter in pipeline_metadata["parameters"]:
            assert "description" in pipeline_metadata["parameters"][parameter]
            assert "type" in pipeline_metadata["parameters"][parameter]

            assert len(pipeline_metadata["parameters"][parameter]) <= 3

            if len(pipeline_metadata["parameters"][parameter]) == 3:
                assert "default" in pipeline_metadata["parameters"][parameter]


def test_metadata_retrieval_for_unregistered_pipelines():
    name = "this pipeline does not exists and can't be registered"
    pipeline_metadata = training_pipeline_name_to_metadata(name)
    assert pipeline_metadata["training_pipeline"] == name
    assert pipeline_metadata["description"] == "A training pipeline."
    assert pipeline_metadata["parameters"] == {}



"""
Sampler implementation.

Reimplemented starting from: https://github.com/ncullen93/torchsample/blob/ea4d1b3975f68be0521941e733887ed667a1b46e/torchsample/samplers.py.
The main reason for reimplementation is to avoid to add a dependency and to control better the logger.
"""

import logging
from typing import Iterator

import numpy as np
import torch
from sklearn.model_selection import StratifiedShuffleSplit
from torch.utils.data import Sampler

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class StratifiedSampler(Sampler):
    """Implementation of a sampler for tensors based on scikit-learn StratifiedShuffleSplit."""

    def __init__(
        self, targets: torch.Tensor, batch_size: int, test_size: float = 0.5
    ) -> None:
        """Construct a StratifiedSampler.

        Args:
            targets: targets tensor.
            batch_size: size of the batch.
            test_size: proportion of samples in the test set. Defaults to 0.5.
        """
        self.targets = targets
        self.number_of_splits = int(self.targets.size(0) / batch_size)
        self.test_size = test_size

    def gen_sample_array(self) -> np.ndarray:
        """Get sample array.

        Returns:
            sample array.
        """
        splitter = StratifiedShuffleSplit(
            n_splits=self.number_of_splits, test_size=self.test_size
        )
        data_placeholder = torch.randn(self.targets.size(0), 2).numpy()
        targets = self.targets.numpy()
        splitter.get_n_splits(data_placeholder, targets)
        train_index, test_index = next(splitter.split(data_placeholder, targets))
        return np.hstack([train_index, test_index])

    def __iter__(self) -> Iterator[np.ndarray]:
        """Get an iterator over the sample array.

        Returns:
            sample array iterator.

        Yields:
            a sample array.
        """
        return iter(self.gen_sample_array())

    def __len__(self) -> int:
        """Length of the sampler.

        Returns:
            the sampler length.
        """
        return len(self.targets)



"""Enzeptional processing tests."""

from gt4sd.frameworks.enzeptional.processing import (
    reconstruct_sequence_with_mutation_range,
    sanitize_intervals,
)


def test_sanitize_intervals():
    assert sanitize_intervals([(-5, 12), (13, 14), (2, 3), (-3, 4), (-2, 6)]) == [
        (-5, 12),
        (13, 14),
    ]


def test_reconstruct_sequence_with_mutation_range():
    assert (
        reconstruct_sequence_with_mutation_range(
            "ABCDEFGHILMNOPQRSTUVWXYZ", "12789", [(0, 1), (6, 8)]
        )
        == "12CDEF789LMNOPQRSTUVWXYZ"
    )



"""Core training utilities."""

from dataclasses import dataclass


class TrainingPipeline:
    """Abstract interface for a training pipelines."""

    def train(self, **kwargs) -> None:
        """Train the models associated to a pipeline."""
        raise NotImplementedError("Can't train an abstract training pipeline.")


@dataclass
class TrainingPipelineArguments:
    """Abstract interface for training pipeline arguments."""

    __name__ = "training_args"



"""HuggingFaceGenerationAlgorithm initialization."""

from .core import (
    HuggingFaceCTRLGenerator,
    HuggingFaceGenerationAlgorithm,
    HuggingFaceGPT2Generator,
    HuggingFaceOpenAIGPTGenerator,
    HuggingFaceTransfoXLGenerator,
    HuggingFaceXLMGenerator,
    HuggingFaceXLNetGenerator,
)

__all__ = [
    "HuggingFaceGenerationAlgorithm",
    "HuggingFaceXLMGenerator",
    "HuggingFaceCTRLGenerator",
    "HuggingFaceGPT2Generator",
    "HuggingFaceOpenAIGPTGenerator",
    "HuggingFaceXLNetGenerator",
    "HuggingFaceTransfoXLGenerator",
]



"""Tests for granular tokenizer."""

from gt4sd.frameworks.granular.tokenizer.tokenizer import (
    SelfiesTokenizer,
    SmilesTokenizer,
)


def test_tokenization():
    smiles = [
        "c1ccccc1",
        "c1ccc(CP(c2ccccc2)c2ccccc2)cc1.CCCCN1[C]N(Cc2ccccc2)c2ccccc21.[Ag]",
    ]

    def _test_tokenizer(tokenizer_type, tokens_groundtruth):
        tokenizer = tokenizer_type("test", smiles=smiles)
        tokens = tokenizer.tokenize(smiles[1])
        assert tokens_groundtruth == tokens
        assert [
            tokenizer.vocab[token] for token in tokens
        ] == tokenizer.convert_tokens_to_ids(tokenizer.tokenize(smiles[1]))
        assert 2 == len(
            [
                tokenizer.convert_tokens_to_ids(tokenizer.tokenize(a_smiles))
                for a_smiles in smiles
            ]
        )

    _test_tokenizer(
        SelfiesTokenizer,
        [
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Branch1_3]",
            "[=S]",
            "[C]",
            "[P]",
            "[Branch1_3]",
            "[Branch2_2]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[Branch1_1]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[Branch1_1]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[#C]",
            "[.]",
            "[C]",
            "[C]",
            "[C]",
            "[C]",
            "[N]",
            "[Cexpl]",
            "[N]",
            "[Branch1_3]",
            "[Branch2_3]",
            "[C]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[Branch1_1]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[c]",
            "[Ring1]",
            "[Branch1_1]",
            "[Ring1]",
            "[=N]",
            "[.]",
            "[Agexpl]",
        ],
    )

    _test_tokenizer(
        SmilesTokenizer,
        [
            "c",
            "1",
            "c",
            "c",
            "c",
            "(",
            "C",
            "P",
            "(",
            "c",
            "2",
            "c",
            "c",
            "c",
            "c",
            "c",
            "2",
            ")",
            "c",
            "2",
            "c",
            "c",
            "c",
            "c",
            "c",
            "2",
            ")",
            "c",
            "c",
            "1",
            ".",
            "C",
            "C",
            "C",
            "C",
            "N",
            "1",
            "[C]",
            "N",
            "(",
            "C",
            "c",
            "2",
            "c",
            "c",
            "c",
            "c",
            "c",
            "2",
            ")",
            "c",
            "2",
            "c",
            "c",
            "c",
            "c",
            "c",
            "2",
            "1",
            ".",
            "[Ag]",
        ],
    )



"""MolGX initialization."""
import logging

from ....extras import EXTRAS_ENABLED

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if EXTRAS_ENABLED:
    from .core import MolGX, MolGXQM9Generator

    __all__ = [
        "MolGX",
        "MolGXQM9Generator",
    ]
else:
    logger.warning("install AMD_analytcs extras to use MolGX")



"""CLaSS tests."""

from typing import ClassVar, Type

import pytest

from gt4sd.algorithms.controlled_sampling.paccmann_gp import (
    PaccMannGP,
    PaccMannGPGenerator,
)
from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.registry import ApplicationsRegistry

TARGET = {
    "qed": {"weight": 1.0},
    "molwt": {"target": 200},
    "sa": {"weight": 2.0},
    "affinity": {"protein": "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT"},
}
PARAMETERS = {
    "number_of_steps": 8,
    "number_of_initial_points": 4,
    "number_of_optimization_rounds": 1,
    "samples_for_evaluation": 2,
    "maximum_number_of_sampling_steps": 4,
}


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            PaccMannGPGenerator,
            "controlled_sampling",
            "materials",
            PaccMannGP.__name__,
        ),
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [
        (PaccMannGPGenerator),
    ],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [
        (PaccMannGPGenerator),
    ],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert "v0" in versions


@pytest.mark.parametrize(
    "config, algorithm, algorithm_parameters",
    [
        (PaccMannGPGenerator, PaccMannGP, PARAMETERS),
    ],
)
def test_generation_via_import(config, algorithm, algorithm_parameters):
    parameters = {
        "batch_size": 1,
    }
    for param, value in algorithm_parameters.items():
        parameters[param] = value
    config = config(**parameters)
    algorithm = algorithm(configuration=config, target=TARGET)
    items = list(algorithm.sample(1))
    assert len(items) == 1


@pytest.mark.parametrize(
    "algorithm_application, algorithm_type, domain, algorithm_name, algorithm_parameters",
    [
        (
            PaccMannGPGenerator.__name__,
            "controlled_sampling",
            "materials",
            PaccMannGP.__name__,
            PARAMETERS,
        ),
    ],
)
def test_generation_via_registry(
    algorithm_type,
    domain,
    algorithm_name,
    algorithm_application,
    algorithm_parameters,
):
    parameters = {
        "target": TARGET,
        "algorithm_type": algorithm_type,
        "domain": domain,
        "algorithm_name": algorithm_name,
        "algorithm_application": algorithm_application,
        "batch_size": 1,
    }
    for param, value in algorithm_parameters.items():
        parameters[param] = value
    algorithm = ApplicationsRegistry.get_application_instance(**parameters)
    items = list(algorithm.sample(1))
    assert len(items) == 1



"""Recurrent Neural Networks with Hill Climbing algorithm implementation."""

from guacamol_baselines.smiles_lstm_hc.goal_directed_generation import (
    SmilesRnnDirectedGenerator,
)


class SMILESLSTMHC:
    def __init__(
        self,
        model_path: str,
        smi_file,
        max_len: int,
        n_jobs: int,
        keep_top: int,
        n_epochs: int,
        mols_to_sample: int,
        optimize_n_epochs: int,
        benchmark_num_samples: int,
        optimize_batch_size: int,
        random_start: bool,
    ):
        """Initialize SMILESLSTMHC.

        Args:
            model_path: path to load the model.
            smi_file: path to load the hypothesis, candidate labels and, optionally, the smiles file.
            max_len: maximum length of a SMILES string.
            n_jobs: number of concurrently running jobs.
            keep_top: molecules kept each step.
            n_epochs: number of epochs to sample.
            mols_to_sample: molecules sampled at each step.
            optimize_n_epochs: number of epochs for the optimization.
            benchmark_num_samples: number of molecules to generate from final model for the benchmark.
            optimize_batch_size: batch size for the optimization.
            random_start: set to True to randomly choose list of SMILES for generating optimized molecules.
        """
        self.model_path = model_path
        self.n_epochs = n_epochs
        self.mols_to_sample = mols_to_sample
        self.keep_top = keep_top
        self.optimize_n_epochs = optimize_n_epochs
        self.max_len = max_len
        self.optimize_batch_size = optimize_batch_size
        self.benchmark_num_samples = benchmark_num_samples
        self.random_start = random_start
        self.smi_file = smi_file
        self.n_jobs = n_jobs

    def get_generator(self) -> SmilesRnnDirectedGenerator:
        """Create an instance of the SmilesRnnDirectedGenerator.

        Returns:
            an instance of SmilesRnnDirectedGenerator.
        """
        optimiser = SmilesRnnDirectedGenerator(
            pretrained_model_path=self.model_path,
            n_epochs=self.n_epochs,
            mols_to_sample=self.mols_to_sample,
            keep_top=self.keep_top,
            optimize_n_epochs=self.optimize_n_epochs,
            max_len=self.max_len,
            optimize_batch_size=self.optimize_batch_size,
            number_final_samples=self.benchmark_num_samples,
            random_start=self.random_start,
            smi_file=self.smi_file,
            n_jobs=self.n_jobs,
        )
        return optimiser



"""GuacaMol algorithms implementation module."""

import logging
import os
from typing import Any, List

from guacamol_baselines.graph_ga.goal_directed_generation import GB_GA_Generator
from guacamol_baselines.graph_mcts.goal_directed_generation import GB_MCTS_Generator
from guacamol_baselines.moses_baselines.aae_distribution_learning import AaeGenerator
from guacamol_baselines.moses_baselines.organ_distribution_learning import (
    OrganGenerator,
)
from guacamol_baselines.moses_baselines.vae_distribution_learning import VaeGenerator
from guacamol_baselines.smiles_ga.goal_directed_generation import ChemGEGenerator
from guacamol_baselines.smiles_lstm_hc.goal_directed_generation import (
    SmilesRnnDirectedGenerator,
)
from guacamol_baselines.smiles_lstm_ppo.goal_directed_generation import (
    PPODirectedGenerator,
)

from .....domains.materials.scorer import CombinedScorer, get_target_parameters
from .....frameworks.torch import claim_device_name
from .graph_ga import GraphGA
from .graph_mcts import GraphMCTS
from .moses_aae import AAE
from .moses_organ import Organ
from .moses_vae import VAE
from .smiles_ga import SMILESGA
from .smiles_lstm_hc import SMILESLSTMHC
from .smiles_lstm_ppo import SMILESLSTMPPO

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Generator:
    """Abstract interface for a conditional generator."""

    def generate_batch(self, target) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        raise NotImplementedError(
            "Implementation not found for generation of molecules."
        )


class SMILESGAIterator(Generator):
    def __init__(
        self,
        resource_path,
        batch_size: int,
        population_size: int,
        n_mutations: int,
        n_jobs: int,
        random_start: bool,
        gene_size: int,
        generations: int,
        patience: int,
    ):
        """Initialize SMILESGAIterator.

        Args:
            resource_path: path to load the hypothesis, candidate labels and, optionally, the smiles file.
            batch_size: number of molecules to generate.
            population_size: used with n_mutations for the initial generation of smiles within the population.
            n_mutations: used with population size for the initial generation of smiles within the population.
            n_jobs: number of concurrently running jobs.
            random_start: set to True to randomly choose list of SMILES for generating optimizied molecules.
            gene_size: size of the gene which is used in creation of genes.
            generations: number of evolutionary generations.
            patience: used for early stopping if population scores remains the same after generating molecules.
        """
        self.resource_path = resource_path
        self.batch_size = batch_size
        self.population_size = population_size
        self.n_mutations = n_mutations
        self.n_jobs = n_jobs
        self.random_start = random_start
        self.gene_size = gene_size
        self.generations = generations
        self.patience = patience
        self.chemGenerator: ChemGEGenerator = None

    def generate_batch(self, target) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        score_list, weights = get_target_parameters(target)

        self.scoring_function = CombinedScorer(
            scorer_list=score_list,
            weights=weights,
        )
        if self.chemGenerator is None:
            optimiser = SMILESGA(
                smi_file=os.path.join(self.resource_path, "guacamol_v1_all.smiles"),
                population_size=self.population_size,
                n_mutations=self.n_mutations,
                gene_size=self.gene_size,
                generations=self.generations,
                n_jobs=self.n_jobs,
                random_start=self.random_start,
                patience=self.patience,
            )
            logger.info("Initialization of the Generator")
            self.chemGenerator = optimiser.get_generator()

        logger.info("generating molecules")
        molecules = self.chemGenerator.generate_optimized_molecules(
            self.scoring_function, self.batch_size
        )
        return molecules


class GraphGAIterator(Generator):
    def __init__(
        self,
        resource_path,
        batch_size: int,
        population_size: int,
        offspring_size: int,
        n_jobs: int,
        mutation_rate: float,
        random_start: bool,
        generations: int,
        patience: int,
    ):
        """Initialize GraphGAIterator.

        Args:
            resource_path: path to load the hypothesis, candidate labels and, optionally, the smiles file.
            batch_size: number of molecules to generate.
            population_size: used for the initial generation of smiles within the population.
            n_jobs: number of concurrently running jobs.
            random_start: set to True to randomly choose list of SMILES for generating optimizied molecules.
            offspring_size: number of molecules to select for new population.
            mutation_rate: frequency of the new mutations in a single gene or organism over time.
            generations: number of evolutionary generations.
            patience: used for early stopping if population scores remains the same after generating molecules.
        """
        self.resource_path = resource_path
        self.batch_size = batch_size
        self.population_size = population_size
        self.n_jobs = n_jobs
        self.random_start = random_start
        self.offspring_size = offspring_size
        self.mutation_rate = mutation_rate
        self.generations = generations
        self.patience = patience
        self.gb_ga_generator: GB_GA_Generator = None

    def generate_batch(self, target) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        score_list, weights = get_target_parameters(target)

        self.scoring_function = CombinedScorer(
            scorer_list=score_list,
            weights=weights,
        )
        if self.gb_ga_generator is None:
            optimiser = GraphGA(
                smi_file=os.path.join(self.resource_path, "guacamol_v1_all.smiles"),
                population_size=self.population_size,
                mutation_rate=self.mutation_rate,
                offspring_size=self.offspring_size,
                generations=self.generations,
                n_jobs=self.n_jobs,
                random_start=self.random_start,
                patience=self.patience,
            )
            logger.info("Initialization of the Generator")
            self.gb_ga_generator = optimiser.get_generator()

        logger.info("generating molecules")
        molecules = self.gb_ga_generator.generate_optimized_molecules(
            self.scoring_function, self.batch_size
        )
        return molecules


class GraphMCTSIterator(Generator):
    def __init__(
        self,
        init_smiles: str,
        batch_size: int,
        population_size: int,
        max_children: int,
        n_jobs: int,
        num_sims: float,
        max_atoms: int,
        generations: int,
        patience: int,
    ):
        """Initialize GraphMCTSIterator.

        Args:
            init_smiles: path where to load hypothesis, candidate labels and, optionally, the smiles file.
            batch_size: number of molecules to generate.
            population_size: used for the initial generation of smiles within the population.
            max_children: maximum number of childerns a node could have.
            n_jobs: number of concurrently running jobs.
            num_sims: number of times to traverse the tree.
            max_atoms: maximum number of atoms to explore to terminal the node state.
            generations: number of evolutionary generations.
            patience: used for early stopping if population scores remains the same after generating molecules.
        """
        self.init_smiles = init_smiles
        self.batch_size = batch_size
        self.population_size = population_size
        self.max_children = max_children
        self.n_jobs = n_jobs
        self.num_sims = num_sims
        self.max_atoms = max_atoms
        self.generations = generations
        self.patience = patience
        self.grah_mcts_generator: GB_MCTS_Generator = None

    def generate_batch(self, target) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        score_list, weights = get_target_parameters(target)

        self.scoring_function = CombinedScorer(
            scorer_list=score_list,
            weights=weights,
        )
        if self.grah_mcts_generator is None:
            optimiser = GraphMCTS(
                init_smiles=self.init_smiles,
                population_size=self.population_size,
                max_children=self.max_children,
                num_sims=self.num_sims,
                generations=self.generations,
                n_jobs=self.n_jobs,
                max_atoms=self.max_atoms,
                patience=self.patience,
            )
            logger.info("Initialization of the Generator")
            self.grah_mcts_generator = optimiser.get_generator()

        logger.info("generating molecules")
        molecules = self.grah_mcts_generator.generate_optimized_molecules(
            self.scoring_function, self.batch_size
        )
        return molecules


class SMILESLSTMHCIterator(Generator):
    def __init__(
        self,
        resource_path,
        batch_size: int,
        n_epochs: int,
        mols_to_sample: int,
        n_jobs: int,
        random_start: bool,
        optimize_n_epochs: int,
        benchmark_num_samples: int,
        keep_top: int,
        max_len: int,
        optimize_batch_size: int,
    ):
        """Initialize SMILESLSTMHCIterator.

        Args:
            resource_path: path to load the hypothesis, candidate labels and, optionally, the smiles file.
            batch_size: number of molecules to generate.
            n_epochs: number of epochs to sample.
            mols_to_sample: molecules sampled at each step.
            keep_top: molecules kept each step.
            optimize_n_epochs: number of epochs for the optimization.
            benchmark_num_samples: number of molecules to generate from final model for the benchmark.
            random_start: set to True to randomly choose list of SMILES for generating optimizied molecules.
            n_jobs: number of concurrently running jobs.
            max_len: maximum length of a SMILES string.
            optimize_batch_size: batch size for the optimization.
        """
        self.resource_path = resource_path
        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.mols_to_sample = mols_to_sample
        self.keep_top = keep_top
        self.optimize_n_epochs = optimize_n_epochs
        self.benchmark_num_samples = benchmark_num_samples
        self.random_start = random_start
        self.n_jobs = n_jobs
        self.max_len = max_len
        self.optimize_batch_size = optimize_batch_size
        self.smiles_lstm_hc_generator: SmilesRnnDirectedGenerator = None

    def generate_batch(self, target) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        score_list, weights = get_target_parameters(target)

        self.scoring_function = CombinedScorer(
            scorer_list=score_list,
            weights=weights,
        )
        if self.smiles_lstm_hc_generator is None:
            optimiser = SMILESLSTMHC(
                model_path=os.path.join(self.resource_path, "model_final_0.473.pt"),
                smi_file=os.path.join(self.resource_path, "guacamol_v1_all.smiles"),
                n_epochs=self.n_epochs,
                mols_to_sample=self.mols_to_sample,
                keep_top=self.keep_top,
                optimize_n_epochs=self.optimize_n_epochs,
                max_len=self.max_len,
                optimize_batch_size=self.optimize_batch_size,
                benchmark_num_samples=self.benchmark_num_samples,
                random_start=self.random_start,
                n_jobs=self.n_jobs,
            )
            logger.info("Initialization of the Generator")
            self.smiles_lstm_hc_generator = optimiser.get_generator()

        logger.info("generating molecules")
        molecules = self.smiles_lstm_hc_generator.generate_optimized_molecules(
            self.scoring_function, self.batch_size
        )
        return molecules


class SMILESLSTMPPOIterator(Generator):
    def __init__(
        self,
        resource_path,
        batch_size: int,
        episode_size: int,
        num_epochs: int,
        optimize_batch_size: int,
        entropy_weight: int,
        kl_div_weight: int,
        clip_param: float,
    ):
        """Initialize SMILESLSTMPPOIterator.

        Args:
            resource_path: path to load the hypothesis, candidate labels and, optionally, the smiles file.
            batch_size: number of molecules to generate.
            episode_size: number of molecules sampled by the policy at the start of a series of ppo updates.
            num_epochs: number of epochs to sample.
            optimize_batch_size: batch size for the optimization.
            entropy_weight: used for calculating entropy loss.
            kl_div_weight: used for calculating Kullback-Leibler divergence loss.
            clip_param: used for determining how far the new policy is from the old one.
        """
        self.resource_path = resource_path
        self.batch_size = batch_size
        self.episode_size = episode_size
        self.num_epochs = num_epochs
        self.optimize_batch_size = optimize_batch_size
        self.entropy_weight = entropy_weight
        self.kl_div_weight = kl_div_weight
        self.clip_param = clip_param
        self.smiles_lstm_ppo_generator: PPODirectedGenerator = None

    def generate_batch(self, target) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        score_list, weights = get_target_parameters(target)

        self.scoring_function = CombinedScorer(
            scorer_list=score_list,
            weights=weights,
        )
        if self.smiles_lstm_ppo_generator is None:
            optimiser = SMILESLSTMPPO(
                model_path=os.path.join(self.resource_path, "model_final_0.473.pt"),
                num_epochs=self.num_epochs,
                episode_size=self.episode_size,
                optimize_batch_size=self.optimize_batch_size,
                entropy_weight=self.entropy_weight,
                kl_div_weight=self.kl_div_weight,
                clip_param=self.clip_param,
            )
            logger.info("initialization of the generator")
            self.smiles_lstm_ppo_generator = optimiser.get_generator()

        logger.info("generating molecules")
        molecules = self.smiles_lstm_ppo_generator.generate_optimized_molecules(
            self.scoring_function, self.batch_size
        )
        return molecules


class AaeIterator:
    def __init__(
        self,
        resource_path: str,
        n_samples: int,
        n_batch: int,
        max_len: int,
    ):
        """Initialize AAE.

        Args:
            resource_path: path to load the hypothesis, candidate labels and, optionally, the smiles file.
            n_samples: number of samples to sample.
            n_batch: size of the batch.
            max_len: max length of SMILES.
        """
        self.resource_path = resource_path
        self.model_path = os.path.join(self.resource_path, "model.pt")
        self.config_path = os.path.join(self.resource_path, "config.pt")
        self.vocab_path = os.path.join(self.resource_path, "vocab.pt")
        self.n_samples = n_samples
        self.n_batch = n_batch
        self.max_len = max_len
        self.aae_generator: AaeGenerator = None
        self.device_name = claim_device_name()

    def generate_batch(self, target=None) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        if self.aae_generator is None:
            optimiser = AAE(
                model_path=self.model_path,
                model_config_path=self.config_path,
                vocab_path=self.vocab_path,
                n_samples=self.n_samples,
                n_batch=self.n_batch,
                max_len=self.max_len,
                device=self.device_name,
            )
            logger.info("Initialization of the Generator")
            self.aae_generator = optimiser.get_generator()
        molecules = self.aae_generator.generate(self.n_samples)
        return molecules


class VaeIterator:
    def __init__(
        self,
        resource_path: str,
        n_samples: int,
        n_batch: int,
        max_len: int,
    ):
        """Initialize VaeIterator.

        Args:
            resource_path: path to load the hypothesis, candidate labels and, optionally, the smiles file.
            n_samples: number of samples to sample.
            n_batch: size of the batch.
            max_len: max length of SMILES.
        """
        self.resource_path = resource_path
        self.model_path = os.path.join(self.resource_path, "model.pt")
        self.config_path = os.path.join(self.resource_path, "config.pt")
        self.vocab_path = os.path.join(self.resource_path, "vocab.pt")
        self.n_samples = n_samples
        self.n_batch = n_batch
        self.max_len = max_len
        self.vae_generator: VaeGenerator = None
        self.device_name = claim_device_name()

    def generate_batch(self, target=None) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        if self.vae_generator is None:
            optimiser = VAE(
                model_path=self.model_path,
                model_config_path=self.config_path,
                vocab_path=self.vocab_path,
                n_samples=self.n_samples,
                n_batch=self.n_batch,
                max_len=self.max_len,
                device=self.device_name,
            )
            logger.info("Initialization of the Generator")
            self.vae_generator = optimiser.get_generator()
        molecules = self.vae_generator.generate(self.n_samples)
        return molecules


class OrganIterator:
    def __init__(
        self,
        resource_path: str,
        n_samples: int,
        n_batch: int,
        max_len: int,
    ):
        """Initialize OrganIterator.

        Args:
            resource_path: path to load the hypothesis, candidate labels and, optionally, the smiles file.
            n_samples: number of samples to sample.
            n_batch: size of the batch.
            max_len: max length of SMILES.
        """
        self.resource_path = resource_path
        self.model_path = os.path.join(self.resource_path, "model.pt")
        self.config_path = os.path.join(self.resource_path, "config.pt")
        self.vocab_path = os.path.join(self.resource_path, "vocab.pt")
        self.n_samples = n_samples
        self.n_batch = n_batch
        self.max_len = max_len
        self.organ_generator: OrganGenerator = None
        self.device_name = claim_device_name()

    def generate_batch(self, target=None) -> List[Any]:
        """Generate a batch of molecules.

        Args:
            target: condition used for generation.

        Returns:
            the generated molecules.
        """
        if self.organ_generator is None:
            optimiser = Organ(
                model_path=self.model_path,
                model_config_path=self.config_path,
                vocab_path=self.vocab_path,
                n_samples=self.n_samples,
                n_batch=self.n_batch,
                max_len=self.max_len,
                device=self.device_name,
            )
            logger.info("Initialization of the Generator")
            self.organ_generator = optimiser.get_generator()
        molecules = self.organ_generator.generate(self.n_samples)
        return molecules



"""Activations for granular models."""

from typing import Dict

from torch import nn

ACTIVATION_FACTORY: Dict[str, nn.Module] = {
    "sigmoid": nn.Sigmoid(),
    "tanh": nn.Tanh(),
    "softmax": nn.Softmax(),
    "relu": nn.ReLU(),
}



"""Catalyst design for NCCR project."""

import logging
import os
import re
from typing import List, Union, cast

import torch

from ......domains.materials import validate_molecules
from ......frameworks.granular.ml.models import (
    GranularEncoderDecoderModel,
    MlpPredictor,
)
from ......frameworks.granular.ml.module import GranularModule
from ......frameworks.granular.tokenizer.tokenizer import SmilesTokenizer
from ..core import (
    GaussianProcessRepresentationsSampler,
    Generator,
    Point,
    PropertyPredictor,
    Representation,
    point_to_tensor,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class CatalystVAE(Representation):
    """Catalyst VAE for suzuki reactions."""

    model: GranularEncoderDecoderModel

    def __init__(
        self,
        resources_path: str,
        padding_length: int = 127,
        maximum_length: int = 100,
        primer_smiles: str = "",
        checkpoint_filename: str = "epoch=199-step=5799.ckpt",
    ) -> None:
        """Constructs a CatalystVAE.

        Args:
            resources_path: directory where to find models and configurations.
            pading_length: size of the padded sequence. Defaults to 127.
            maximum_length: maximum length of the synthesis.
            primer_smiles: primer SMILES representation. Default to "", a.k.a., no primer.
            checkpoint_filename: checkpoint filename. Defaults to "epoch=199-step=5799.ckpt".
        """
        self.vocabulary_filepath = os.path.join(resources_path, "vocab_combined.csv")
        self.checkpoint_filepath = os.path.join(
            resources_path, "epoch=199-step=5799.ckpt"
        )
        self.tokenizer = SmilesTokenizer(self.vocabulary_filepath)
        self.model = cast(
            GranularEncoderDecoderModel,
            GranularModule.load_from_checkpoint(self.checkpoint_filepath).autoencoders[
                0
            ],
        )
        self.model.eval()
        self.padding_length = padding_length
        self.z_dimension = self.model.latent_size
        self.maximum_length = maximum_length
        self.primer_smiles = primer_smiles
        if len(self.primer_smiles) > 0:
            self.primer_point = self.smiles_to_latent(self.primer_smiles)
        else:
            self.primer_point = torch.zeros(1, self.z_dimension)
        self.clean_regex = re.compile(
            r"{}|{}".format(self.tokenizer.sos_token, self.tokenizer.unk_token)
        )
        self.end_regex = re.compile(
            r"{}|{}".format(self.tokenizer.eos_token, self.tokenizer.pad_token)
        )

    def smiles_to_latent(self, smiles: str) -> Point:
        """Encode a SMILES into a latent point.

        Args:
            smiles: a SMILES representation of a molecule.

        Returns:
            the encoded latent space point.
        """
        return self.model.encode(  # type:ignore
            torch.tensor(
                [
                    self.tokenizer.add_padding_tokens(
                        self.tokenizer.convert_tokens_to_ids(
                            [self.tokenizer.sos_token]
                            + self.tokenizer.tokenize(smiles)
                            + [self.tokenizer.eos_token]
                        ),
                        length=self.padding_length,
                    )
                ]
            )
        )

    def decode(self, z: Point) -> str:
        """Decode a catalyst from the latent space.

        Args:
            z: a latent space point.

        Returns:
            a catalyst in SMILES format.
        """
        z = torch.unsqueeze(point_to_tensor(z), dim=0)
        reconstructed = self.model.decode(z, max_len=self.maximum_length)[0][
            0
        ]  # type:ignore
        reconstructed = self.clean_regex.sub("", reconstructed)
        match_ending = self.end_regex.search(reconstructed)
        if match_ending:
            reconstructed = reconstructed[: match_ending.start()]
        return reconstructed


class CatalystBindingEnergyPredictor(PropertyPredictor):
    """Catalyst binding energy predictor for suzuki reactions."""

    model: MlpPredictor

    def __init__(
        self, resources_path: str, checkpoint_filename: str = "epoch=199-step=5799.ckpt"
    ) -> None:
        """Constructs a CatalystBindingEnergyPredictor.

        Args:
            resources_path: directory where to find models and configurations.
            checkpoint_filename: checkpoint filename. Defaults to "epoch=199-step=5799.ckpt".
        """
        self.vocabulary_filepath = os.path.join(resources_path, "vocab_combined.csv")
        self.checkpoint_filepath = os.path.join(resources_path, checkpoint_filename)
        self.tokenizer = SmilesTokenizer(self.vocabulary_filepath)
        self.model = cast(
            MlpPredictor,
            GranularModule.load_from_checkpoint(self.checkpoint_filepath).latent_models[
                0
            ],
        )
        self.model.eval()

    def __call__(self, z: Point) -> float:
        """Predict binding energy.

        Args:
            z: a latent space point.

        Returns:
            the predicted binding energy.
        """
        z = point_to_tensor(z)
        return self.model(z)[0][0].item()


class CatalystGenerator(Generator):
    """Catalyst generator."""

    def __init__(
        self,
        resources_path: str,
        generated_length: int = 100,
        number_of_points: int = 10,
        number_of_steps: int = 50,
        primer_smiles: str = "",
        checkpoint_filename: str = "epoch=199-step=5799.ckpt",
    ):
        """Constructs catalyst generator.

        Args:
            resource_path: directory where to find models and configurations.
            generated_length: maximum lenght of the generated molecule. Defaults to 100.
            number_of_points: number of optimal points to return. Defaults to 10.
            number_of_steps: number of optimization steps. Defaults to 50.
            primer_smiles: primer SMILES representation. Default to "", a.k.a., no primer.
            checkpoint_filename: checkpoint filename. Defaults to "epoch=199-step=5799.ckpt".
        """
        self.resources_path = resources_path
        self.checkpoint_filename = checkpoint_filename
        self.generated_length = generated_length
        self.number_of_points = number_of_points
        self.number_of_steps = max(self.number_of_points, number_of_steps)
        self.primer_smiles = primer_smiles
        self.vae = CatalystVAE(
            resources_path,
            maximum_length=self.generated_length,
            primer_smiles=primer_smiles,
            checkpoint_filename=checkpoint_filename,
        )
        self.predictor = CatalystBindingEnergyPredictor(
            resources_path, checkpoint_filename=checkpoint_filename
        )
        self.minimum_latent_coordinate = -100.0
        self.maximum_latent_coordinate = 100.0

    def generate_samples(self, target_energy: Union[float, str]) -> List[str]:
        """Generate samples given a target energy.

        Args:
            target_energy: target energy value.

        Returns:
            catalysts sampled for the target value.
        """
        if isinstance(target_energy, str):
            logger.warning(
                f"target energy ({target_energy}) passed as string, casting to float"
            )
            target_energy = float(target_energy)
        sampler = GaussianProcessRepresentationsSampler(
            {"energy": target_energy},
            property_predictors={"energy": self.predictor},
            representations={"smiles": self.vae},
            bounds={
                "smiles": (
                    self.minimum_latent_coordinate,
                    self.maximum_latent_coordinate,
                )
            },
        )
        smiles_list = [
            sample["smiles"]
            for sample in sampler.optimize(
                number_of_points=self.number_of_points,
                number_of_steps=self.number_of_steps,
            )
            if len(sample["smiles"])
        ]
        _, valid_indexes = validate_molecules(smiles_list=smiles_list)
        return [smiles_list[index] for index in valid_indexes]



"""
Implementation details for TorchDrug generation algorithms.

Parts of the implementation inspired by: https://torchdrug.ai/docs/tutorials/generation.html.
"""
import logging
from pathlib import Path
from typing import List, Optional, Union

import torch
from torch import optim
from torchdrug import core, models, tasks
from torchdrug.core.engine import Engine
from torchdrug.layers import distribution
from torchdrug.tasks.generation import AutoregressiveGeneration, GCPNGeneration

# isort: off
from torch import nn

# isort: on

from ....frameworks.torch import device_claim

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

Task = Union[GCPNGeneration, AutoregressiveGeneration]

"""
Necessary because torchdrug silently overwrites the default nn.Module. This is quite
invasive and causes significant side-effects in the rest of the code.
See: https://github.com/DeepGraphLearning/torchdrug/issues/77
"""
nn.Module = nn._Module  # type: ignore


class DummyDataset:
    """A helper class to imitate a torchdrug dataset."""

    def __init__(self, atom_types: List[int]):
        self.atom_types = atom_types
        self.transform = None


class Generator:
    """Implementation of a TorchDrug generator."""

    solver: Engine
    task: Task
    num_sample: int = 32
    max_resample: int = 16

    def __init__(
        self,
        resources_path: str,
        atom_types: List[int],
        hidden_dims: List[int],
        input_dim: int,
        num_relation: int,
        batch_norm: bool,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """A TorchDrug generation algorithm.

        Args:
            resources_path: path to the cache.
            atom_types: list of atom types.
            hidden_dims: list of hidden dimensions, one per layer.
            num_relation: number of relations for the graph.
            batch_norm: whether to use batch normalization.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided,
                it is inferred.
        """
        self.device = device_claim(device)
        self.resources_path = resources_path

        self.model = models.RGCN(
            input_dim=input_dim,
            num_relation=num_relation,
            hidden_dims=hidden_dims,
            batch_norm=batch_norm,
        )

        self.dataset = DummyDataset(atom_types)

    def load_model(self, resources_path: str):
        """Load a pretrained TorchDrug model."""
        self.solver.load(Path(resources_path).joinpath("weights.pkl"))

    def sample(self) -> List[str]:
        """Sample a molecule.

        Returns:
            a generated SMILES string wrapped into a list.
        """
        results = self.task.generate(
            num_sample=self.num_sample, max_resample=self.max_resample
        )
        return results.to_smiles()


class GCPNGenerator(Generator):
    """
    Interface for the GCPN model as implemented in TorchDrug.

    For details see:
    You, J. et al. (2018). Graph convolutional policy network for goal-directed
    molecular graph generation. Advances in neural information processing systems, 31.

    """

    task: GCPNGeneration

    input_dim: int = 18
    num_relation: int = 3
    batch_norm: bool = False
    atom_types: List[int] = [6, 7, 8, 9, 15, 16, 17, 35, 53]
    hidden_dims: List[int] = [256, 256, 256, 256]

    def __init__(self, resources_path: str):
        """
        A TorchDrug Graph-convolutional policy network (GCPN) generation algorithm.

        Args:
            resources_path: path to the cache.
        """

        super().__init__(
            input_dim=self.input_dim,
            num_relation=self.num_relation,
            batch_norm=self.batch_norm,
            atom_types=self.atom_types,
            hidden_dims=self.hidden_dims,
            resources_path=resources_path,
        )

        self.task = tasks.GCPNGeneration(
            self.model,
            self.atom_types,
            max_edge_unroll=12,
            max_node=38,
            criterion="nll",
        )
        optimizer = optim.Adam(self.task.parameters(), lr=1e-3)
        self.solver = core.Engine(self.task, self.dataset, None, None, optimizer)
        self.load_model(resources_path)


class GAFGenerator(Generator):
    """
    Interface for the GraphAF model as implemented in TorchDrug.

    For details see:
    Shi, Chence, et al. "GraphAF: a Flow-based Autoregressive Model for Molecular
    Graph Generation" International Conference on Learning Representations (ICLR), 2020.
    """

    task: AutoregressiveGeneration

    input_dim: int = 9
    num_relations: int = 3
    batch_norm: bool = True
    atom_types: List[int] = [6, 7, 8, 9, 15, 16, 17, 35, 53]
    hidden_dims: List[int] = [256, 256, 256]

    def __init__(self, resources_path: str):
        """
        A TorchDrug flow-based autoregressive graph generation algorithm (GAF).

        Args:
            resources_path (str): path to the cache.
        """
        super().__init__(
            input_dim=self.input_dim,
            num_relation=self.num_relations,
            batch_norm=self.batch_norm,
            atom_types=self.atom_types,
            hidden_dims=self.hidden_dims,
            resources_path=resources_path,
        )

        node_prior = distribution.IndependentGaussian(
            torch.zeros(self.input_dim), torch.ones(self.input_dim)
        )
        edge_prior = distribution.IndependentGaussian(
            torch.zeros(self.num_relations + 1), torch.ones(self.num_relations + 1)
        )
        node_flow = models.GraphAF(self.model, node_prior, num_layer=12)
        edge_flow = models.GraphAF(self.model, edge_prior, use_edge=True, num_layer=12)

        self.task = tasks.AutoregressiveGeneration(
            node_flow, edge_flow, max_node=38, max_edge_unroll=12, criterion="nll"
        )
        optimizer = optim.Adam(self.task.parameters(), lr=1e-3)
        self.solver = core.Engine(self.task, self.dataset, None, None, optimizer)
        self.load_model(resources_path)



"""Prediction algorithms based on PaccMann"""

import logging
from dataclasses import field
from typing import Any, ClassVar, List, Optional, TypeVar

from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import BimodalMCAAffinityPredictor, MCAPredictor

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)


class PaccMann(GeneratorAlgorithm[S, T]):
    """PaccMann predictor."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T] = None,
    ):
        """Instantiate PaccMann for prediction.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for predicting affinity for a given ligand and target protein pair::

                config = AffinityPredictor(
                    protein_targets=[
                        "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
                        "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
                    ],
                    ligands=[
                        "CONN=COc1cc2ccccc2c11Occncc(Cl)c1N(O)O",
                        "ClCCC(O1)C(C(N=C1C(=O)Nc1cccc(F)c1F)SO)C",
                    ]
                )
                algorithm = PaccMann(configuration=config)
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to perform the prediction via PaccMann's generator.

        Args:
            configuration: helps to set up specific application of PaccMann.
            target: context or condition for the generation.

        Returns:
            callable with target predicting properties using PaccMann.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: MCAPredictor = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.predict_values


@ApplicationsRegistry.register_algorithm_application(PaccMann)
class AffinityPredictor(AlgorithmConfiguration[str, str]):
    """Configuration to predict affinity for a given ligand/protrin target pair."""

    algorithm_type: ClassVar[str] = "prediction"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    protein_targets: List[str] = field(
        default_factory=list,
        metadata=dict(description="List of protein targets as AA sequences."),
    )
    ligands: List[str] = field(
        default_factory=list,
        metadata=dict(description="List of ligands in SMILES format."),
    )
    confidence: bool = field(
        default=False,
        metadata=dict(
            description="Whether the confidence for the prediction should be returned."
        ),
    )

    def get_conditional_generator(
        self, resources_path: str
    ) -> BimodalMCAAffinityPredictor:
        """Instantiate the actual predictor implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`gt4sd.algorithms.prediction.affinity._predicto.implementation.BimodalMCAAffinityPredictor.predict` method for predicting affinity.
        """
        return BimodalMCAAffinityPredictor(
            resources_path=resources_path,
            protein_targets=self.protein_targets,
            ligands=self.ligands,
            confidence=self.confidence,
        )



"""Generic modules."""

import copy
import math
from typing import Any, Callable, List, Optional, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

from ....torch import get_device_from_tensor
from ...tokenizer import Tokenizer
from .activation import ACTIVATION_FACTORY


class Mlp(nn.Module):
    """MLP module."""

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        output_size: int,
        n_layers: int,
        activation: str,
        dropout: float,
        **kwargs,
    ) -> None:
        """Construct Mlp.

        Args:
            input_size: size of the input.
            hidden_size: size of the hidden layers.
            output_size: size of the output.
            n_layers: number of layers.
            activation: name of the activation.
            dropout: dropout rate.
        """
        super().__init__()
        activation = activation.lower()
        self.activation = ACTIVATION_FACTORY.get(activation, None)
        self.first_layer = nn.Linear(input_size, hidden_size)
        middle_layers: List[nn.Module] = list()
        for _ in range(n_layers):
            middle_layers.append(nn.Linear(hidden_size, hidden_size))
            middle_layers.append(nn.ReLU())
            middle_layers.append(nn.Dropout(p=dropout))
        self.middle_layers = nn.Sequential(*middle_layers)
        self.last_layer = nn.Linear(hidden_size, output_size)
        self.relu = nn.ReLU()
        self.output_dim = output_size

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: model input.

        Returns:
            model output.
        """
        z = self.first_layer(x)
        z = self.relu(z)
        z = self.middle_layers(z)
        z = self.last_layer(z)
        if self.activation:
            z = self.activation(z)
        return z


class MlpEncoder(Mlp):
    """MLP encoder."""

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        output_size: int,
        n_layers: int,
        activation: str,
        dropout: float = 0.0,
        **kwargs,
    ) -> None:
        """Construct MlpEncoder.

        Args:
            input_size: size of the input.
            hidden_size: size of the hidden layers.
            output_size: size of the output.
            n_layers: number of layers.
            activation: name of the activation.
            dropout: dropout rate. Defaults to 0.0.
        """
        super().__init__(
            input_size=input_size,
            hidden_size=hidden_size,
            output_size=output_size,
            n_layers=n_layers,
            activation=activation,
            dropout=dropout,
        )


class MlpDecoder(Mlp):
    """MLP decoder."""

    def __init__(
        self,
        latent_size: int,
        hidden_size: int,
        output_size: int,
        n_layers: int,
        activation: str,
        dropout: float = 0.0,
        **kwargs,
    ) -> None:
        """Construct MlpEncoder.

        Args:
            latent_size: size of the input.
            hidden_size: size of the hidden layers.
            output_size: size of the output.
            n_layers: number of layers.
            activation: name of the activation.
            dropout: dropout rate. Defaults to 0.0.
        """
        super().__init__(
            input_size=latent_size,
            hidden_size=hidden_size,
            output_size=output_size,
            n_layers=n_layers,
            activation=activation,
            dropout=dropout,
        )


class RnnEncoder(nn.Module):
    """RNN encoder."""

    def __init__(
        self,
        vocab_size: int,
        embedding_size: int,
        hidden_size: int = 256,
        n_layers: int = 2,
        bidirectional: bool = False,
        latent_size: int = 196,
    ) -> None:
        """Construct RnnEncoder.

        Args:
            vocab_size: size of the vocabulary.
            embedding_size: size of the embedding vectors.
            hidden_size: hidden size. Defaults to 256.
            n_layers: number of layers. Defaults to 2.
            bidirectional: whether the RNN cell is bidirectional. Defaults to False.
            latent_size: latent size. Defaults to 196.
        """
        super().__init__()
        self.input_size = embedding_size
        self.hidden_size = hidden_size
        self.bidirectional = bidirectional
        self.latent_size = latent_size
        self.hidden_factor = (2 if bidirectional else 1) * n_layers
        self.rnn = nn.GRU(
            input_size=embedding_size,
            hidden_size=hidden_size,
            num_layers=n_layers,
            bidirectional=bidirectional,
            batch_first=True,
        )
        self.embedding = nn.Embedding(
            num_embeddings=vocab_size, embedding_dim=embedding_size
        )

    def forward(
        self, input_sequence: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass in the model.

        Args:
            input_sequence: input sequence tensor.

        Returns:
            a tuple containing hidden state and embedded sequence.
        """
        input_embedding = self.embedding(input_sequence)
        _, hidden = self.rnn(input_embedding)
        hidden = hidden.permute(1, 0, 2)
        hidden = hidden.contiguous().view(hidden.size(0), -1)
        return hidden, input_embedding


class RnnDecoder(nn.Module):
    """RNN decoder."""

    def __init__(
        self,
        vocab_size: int,
        embedding_size: int,
        hidden_size: int = 256,
        n_layers: int = 2,
        latent_size: int = 196,
    ) -> None:
        """Construct RnnDecoder.

        Args:
            vocab_size: size of the vocabulary.
            embedding_size: size of the embedding vectors.
            hidden_size: hidden size. Defaults to 256.
            n_layers: number of layers. Defaults to 2.
            latent_size: latent size. Defaults to 196.
        """
        super().__init__()

        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.n_layers = n_layers
        self.latent_size = latent_size
        self.hidden_factor = n_layers
        self.rnn = nn.GRU(
            input_size=embedding_size,
            hidden_size=hidden_size,
            num_layers=n_layers,
            batch_first=True,
        )
        self.latent2hidden = torch.nn.Linear(
            latent_size, hidden_size * self.hidden_factor
        )
        self.outputs2vocab = torch.nn.Linear(hidden_size, vocab_size)

    def forward(
        self, latent: torch.Tensor, input_embedding: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            latent: latent tensor.
            input_embedding: input embedding.

        Returns:
            model output.
        """
        hidden = self.latent2hidden(latent)
        hidden = hidden.view(-1, self.hidden_factor, self.hidden_size)
        hidden = hidden.permute(1, 0, 2).contiguous()
        hidden = torch.tanh(hidden)
        outputs, _ = self.rnn(input_embedding, hidden)
        b, seq_len, hsize = outputs.size()
        outputs = outputs.contiguous().view(-1, hsize)
        outputs = self.outputs2vocab(outputs)
        outputs = outputs.view(b, seq_len, self.vocab_size)
        return outputs

    def inference_direct(
        self,
        latent: torch.Tensor,
        embedding: nn.Module,
        tokenizer: Tokenizer,
        max_len: int,
    ) -> Tuple[List[str], torch.Tensor]:
        """Direct inference from latent space.

        Args:
            latent: latent tensor.
            embedding: embedding module.
            tokenizer: tokenizer.
            max_len: maximum sequence length.

        Returns:
            a tuple containing decoded strings and indices.
        """
        batch_size = latent.size(0)
        hidden = self.latent2hidden(latent)
        hidden = hidden.view(batch_size, self.hidden_factor, self.hidden_size)
        hidden = hidden.permute(1, 0, 2).contiguous()
        hidden = torch.tanh(hidden)
        input_sequence = torch.full(
            (batch_size, 1), tokenizer.sos_token_id, device=latent.device
        ).long()
        logits_list = []
        for t in range(max_len):
            input_embedding = embedding(input_sequence)
            output, hidden = self.rnn(input_embedding, hidden)
            logits = self.outputs2vocab(output)
            logits_list.append(logits)
            input_sequence = torch.argmax(logits, dim=-1)

        logits_tensor = torch.cat(logits_list, dim=1)
        token_indices = torch.argmax(logits_tensor, dim=-1)
        decoded_texts = []
        for index in range(batch_size):
            tokens = [
                tokenizer.convert_id_to_token(vocab_index.item())
                for vocab_index in token_indices[index]
            ]
            text = "".join(tokens).split()[0]
            decoded_texts.append(text)
        return decoded_texts, token_indices


def attention(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    mask: Optional[torch.Tensor] = None,
    dropout: Optional[nn.Module] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Compute scaled dot product attention (adapted from Viswani et al.).

    Args:
        query: query tensor.
        key: key tensor.
        value: value tesor.
        mask: mask to apply on attention score. Defaults to None, a.k.a., no mask.
        dropout: dropout layer. Defaults to None, a.k.a., no dropout.

    Returns:
        a tuple containing the applied attention and the attention weights.
    """
    d_k = query.size(-1)
    scores = torch.matmul(query, key.transpose(-2, -1)) / math.sqrt(d_k)
    if mask is not None:
        scores = scores.masked_fill(mask == 0, -1e9)
    p_attn = F.softmax(scores, dim=-1)
    if dropout is not None:
        p_attn = dropout(p_attn)
    return torch.matmul(p_attn, value), p_attn


def clones(module: nn.Module, n: int) -> nn.Module:
    """Produce N identical layers (adapted from http://nlp.seas.harvard.edu/2018/04/03/attention.html).

    Args:
        module: a module.
        n: number of clones.

    Returns:
        a module list.
    """
    return torch.nn.ModuleList([copy.deepcopy(module) for _ in range(n)])


def subsequent_mask(size: int) -> torch.Tensor:
    """Mask out subsequent positions (adapted from http://nlp.seas.harvard.edu/2018/04/03/attention.html).

    Args:
        size: size of the attention matrix.

    Returns:
        the mask tensor.
    """
    attn_shape = (1, size, size)
    subsequent_mask = np.triu(np.ones(attn_shape), k=1).astype("uint8")
    return torch.from_numpy(subsequent_mask) == 0


class ListModule(torch.nn.Module):
    """Create single pytorch module from list of modules."""

    def __init__(self, *args) -> None:
        """Construct ListModule."""
        super().__init__()
        idx = 0
        for module in args:
            self.add_module(str(idx), module)
            idx += 1

    def __getitem__(self, idx: int) -> Any:
        """Get item from the module list.

        Args:
            idx: index of the item.

        Raises:
            IndexError: in case the index is out of range.

        Returns:
            the item.
        """
        if idx < 0 or idx >= len(self._modules):
            raise IndexError("index {} is out of range".format(idx))
        it = iter(self._modules.values())
        for i in range(idx):
            next(it)
        return next(it)

    def __iter__(self) -> Any:
        """An iterator over the module list values.

        Returns:
            the iterator over values.
        """
        return iter(self._modules.values())

    def __len__(self):
        """Length of the module list.

        Returns:
            the number of modules.
        """
        return len(self._modules)


class MultiHeadedAttention(nn.Module):
    """Multihead attention implementation (based on Vaswani et al.)."""

    def __init__(self, h, d_model, dropout=0.1) -> None:
        """Construct MultiHeadedAttention.

        Args:
            h: number of heads.
            d_model: model size.
            dropout: dropout rate. Defaults to 0.1.
        """
        super().__init__()
        assert d_model % h == 0
        # we assume d_v always equals d_k
        self.d_k = d_model // h
        self.h = h
        self.linears = clones(nn.Linear(d_model, d_model), 4)
        self.attn = None
        self.dropout = nn.Dropout(p=dropout)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        return_attn: bool = False,
    ) -> Any:
        """Forward pass in the model.

        Args:
            query: query tensor.
            key: key tensor.
            value: value tesor.
            mask: mask to apply on attention score. Defaults to None, a.k.a., no mask.
            return_attn: whether to return the attention matrix instead of the linear layer output.
                Defaults to False, a.k.a, do not return attention.

        Returns:
            either the last layer output of the attention matrix.
        """
        if mask is not None:
            # Same mask applied to all h heads
            mask = mask.unsqueeze(1)
        nbatches = query.size(0)

        # 1) do all the linear projections in batch from d_model => h x d_k
        query, key, value = [
            linear_layer(x).view(nbatches, -1, self.h, self.d_k).transpose(1, 2)
            for linear_layer, x in zip(self.linears, (query, key, value))  # type:ignore
        ]

        # 2) apply attention on all the projected vectors in batch
        x, self.attn = attention(  # type:ignore
            query, key, value, mask=mask, dropout=self.dropout
        )  # type:ignore

        # 3) "concat" using a view and apply a final linear
        x = x.transpose(1, 2).contiguous().view(nbatches, -1, self.h * self.d_k)
        if return_attn:
            return self.attn
        else:
            return self.linears[-1](x)  # type:ignore


class PositionwiseFeedForward(nn.Module):
    """Feed forward implementation."""

    def __init__(self, d_model: int, d_ff: int, dropout: float = 0.1) -> None:
        """Construct PositionwiseFeedForward.

        Args:
            d_model: model size.
            d_ff: feed forward size.
            dropout: dropout rate. Defaults to 0.1.
        """
        super().__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.

        Returns:
            feed forward output.
        """
        return self.w_2(self.dropout(F.relu(self.w_1(x))))


class ConvBottleneck(nn.Module):
    """Set of convolutional layers to reduce memory matrix to single latent vector."""

    def __init__(self, size: int, number_of_layers: int = 3) -> None:
        """Construct ConvBottleneck.

        Args:
            size: input size.
            number_of_layers: convolutional layers number. Defaults to 3.
        """
        super().__init__()
        conv_layers = []
        in_d = size
        first = True
        for i in range(number_of_layers):
            out_d = int((in_d - 64) // 2 + 64)
            if first:
                kernel_size = 9
                first = False
            else:
                kernel_size = 8
            if i == 2:
                out_d = 64
            conv_layers.append(
                nn.Sequential(nn.Conv1d(in_d, out_d, kernel_size), nn.MaxPool1d(2))
            )
            in_d = out_d
        self.conv_layers = ListModule(*conv_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.

        Returns:
            model output.
        """
        for conv in self.conv_layers:
            x = F.relu(conv(x))
        return x


class DeconvBottleneck(nn.Module):
    """Set of deconvolutional layers to reshape latent vector back into memory matrix."""

    def __init__(self, size: int, seq_len: int, dim_factor: int) -> None:
        """Construct DeconvBottleneck.

        Args:
            size: size of the deconvolutional padding.
            seq_len: length of the sequence.
            dim_factor: dimensionality factor.
        """
        super().__init__()
        deconv_layers = []

        in_d = 64

        out_fac = 9 * dim_factor + 8
        out_fac = out_fac - 1 + 50 + 1
        diff_seq = out_fac - seq_len

        for i in range(3):
            out_d = (size - in_d) // 4 + in_d
            stride = 3
            padding = 3
            dilation = 1
            kernel_size = 11
            output_padding = 0
            if i == 2:
                out_d = size
                stride = 1
                dilation = 5
                if diff_seq % 2 == 0:
                    padding = int(diff_seq / 2)
                    output_padding = 0
                else:
                    padding = math.ceil(diff_seq / 2)
                    output_padding = 1

            deconv_layers.append(
                nn.Sequential(
                    nn.ConvTranspose1d(
                        in_d,
                        out_d,
                        kernel_size,
                        dilation=dilation,
                        stride=stride,
                        padding=padding,
                        output_padding=output_padding,
                    )
                )
            )
            in_d = out_d
        self.deconv_layers = ListModule(*deconv_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.

        Returns:
            model output.
        """
        for deconv in self.deconv_layers:
            x = F.relu(deconv(x))
        return x


class Embeddings(nn.Module):
    "Transforms input token id tensors to size d_model embeddings."

    def __init__(self, d_model: int, vocab_size: int) -> None:
        """Costruct Embeddings.

        Args:
            d_model: size of the embedding vectors.
            vocab_size: size of the vocabulary.
        """
        super().__init__()
        self.lut = nn.Embedding(vocab_size, d_model)
        self.d_model = d_model

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.

        Returns:
            model output.
        """
        return self.lut(x) * math.sqrt(self.d_model)


class PositionalEncoding(nn.Module):
    """Static sinusoidal positional encoding layer."""

    def __init__(self, d_model: int, dropout: float, max_len: int = 5000) -> None:
        """Construct PositionalEncoding.

        Args:
            d_model: model size.
            dropout: dropout rate.
            max_len: maximum sequence length. Defaults to 5000.
        """
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        # compute the positional encodings once in log space
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * -(math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.

        Returns:
            model output.
        """
        x = x + torch.autograd.Variable(
            self.pe[:, : x.size(1)], requires_grad=False  # type:ignore
        )
        return self.dropout(x)


class TorchLayerNorm(nn.Module):
    """Layer normalization using torch BatchNorm1d."""

    def __init__(self, features: int, eps=1e-6) -> None:
        """Construct TorchLayerNorm.

        Args:
            features: number of features.
            eps: espilon to add to denominator for numerical stability. Defaults to 1e-6.
        """
        super().__init__()
        self.bn = nn.BatchNorm1d(features, eps=eps)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.

        Returns:
            model output.
        """
        return self.bn(x)


class LayerNorm(nn.Module):
    """Custom layer normalization."""

    def __init__(self, features: int, eps=1e-6) -> None:
        """Construct LayerNorm.

        Args:
            features: number of features.
            eps: espilon to add to denominator for numerical stability. Defaults to 1e-6.
        """
        super().__init__()
        self.a = nn.Parameter(torch.ones(features))
        self.b = nn.Parameter(torch.zeros(features))
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.

        Returns:
            model output.
        """
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.a * (x - mean) / (std + self.eps) + self.b


class SublayerConnection(nn.Module):
    """A residual connection followed by a layer normalization.

    Note for code simplicity the norm is first as opposed to last. A dropout layer
    is also applied.
    """

    def __init__(self, size: int, dropout: float) -> None:
        super().__init__()
        self.norm = LayerNorm(size)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor, sublayer: Callable) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.
            sublayer: a callable returning a tensor.

        Returns:
            model output.
        """
        return x + self.dropout(sublayer(self.norm(x)))


class TransformerEncoder(nn.Module):
    """Base transformer encoder architecture."""

    def __init__(
        self,
        hidden_size: int,
        ff_size: int,
        seq_len: int,
        dropout: float,
        heads: int,
        n_layers_enc: int,
        vocab_size: int,
        bypass_bottleneck: bool,
    ) -> None:
        """Construct TransformerEncoder.

        Args:
            hidden_size: hidden size.
            ff_size: feed forward size.
            seq_len: sequence length.
            dropout: dropout rate.
            heads: number of heads.
            n_layers_enc: number of encoding layers.
            vocab_size: vocabulary size.
            bypass_bottleneck: whether the bottleneck should be by passed.
        """
        super().__init__()

        self.position = PositionalEncoding(hidden_size, dropout)
        self.embedding = nn.Sequential(
            Embeddings(hidden_size, vocab_size * 2), self.position
        )

        self.self_attn = MultiHeadedAttention(heads, hidden_size)
        self.feed_forward = PositionwiseFeedForward(hidden_size, ff_size, dropout)
        layer = TransformerEncoderLayer(
            hidden_size, seq_len, self.self_attn, self.feed_forward, dropout
        )
        self.layers = clones(layer, n_layers_enc)

        self.conv_bottleneck = ConvBottleneck(hidden_size)
        self.norm = LayerNorm(hidden_size)

        self.bypass_bottleneck = bypass_bottleneck
        conv_output_shape = self.calc_output_shape(seq_len, hidden_size)
        self.conv_output_len = conv_output_shape[1] * conv_output_shape[2]
        self.conv_output_shape = conv_output_shape

    def calc_output_shape(self, seq_len: int, hidden_size: int):
        """Compute output shape.

        Args:
            seq_len: sequence length.
            hidden_size: hidden size.

        Returns:
            convolutional bottleneck output shape.
        """
        x = torch.randn((1, hidden_size, seq_len))
        x_out = self.conv_bottleneck(x)
        return x_out.shape

    def forward(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.
            mask: mask to apply in the attention layer.

        Returns:
            model output.
        """
        x = self.embedding(x)
        for _, attn_layer in enumerate(self.layers):  # type:ignore
            x = attn_layer(x, mask)
        mem = self.norm(x)
        mem = mem.permute(0, 2, 1)
        mem = self.conv_bottleneck(mem)
        mem = mem.contiguous().view(mem.size(0), -1)
        return mem


class TransformerEncoderLayer(nn.Module):
    """Self-attention/feedforward implementation."""

    def __init__(
        self,
        size: int,
        seq_len: int,
        self_attn: nn.Module,
        feed_forward: nn.Module,
        dropout: float,
    ) -> None:
        """Construct TransformerEncoderLayer.

        Args:
            size: model size.
            seq_len: sequence length.
            self_attn: self-attention layer.
            feed_forward: feed forward layer.
            dropout: droupout rate.
        """
        super().__init__()
        self.size = size
        self.seq_len = seq_len
        self.self_attn = self_attn
        self.feed_forward = feed_forward
        self.sublayer = clones(SublayerConnection(self.size, dropout), 2)

    def forward(
        self, x: torch.Tensor, mask: torch.Tensor, return_attn: bool = False
    ) -> Any:
        """Forward pass in the model.

        Args:
            x: input tensor.
            mask: mask to apply in the attention layer.
            return_attn: whether to return the attention together with the output.
                Defaults to False, return only encoder output.

        Returns:
            model output.
        """
        if return_attn:
            attn = self.self_attn(x, x, x, mask, return_attn=True)
            x = self.sublayer[0](  # type:ignore
                x, lambda x: self.self_attn(x, x, x, mask)
            )
            return self.sublayer[1](x, self.feed_forward), attn  # type:ignore
        else:
            x = self.sublayer[0](  # type:ignore
                x, lambda x: self.self_attn(x, x, x, mask)
            )
            return self.sublayer[1](x, self.feed_forward)  # type:ignore


class TransformerDecoder(nn.Module):
    """Base transformer decoder architecture."""

    def __init__(
        self,
        hidden_size: int,
        ff_size: int,
        seq_len: int,
        dropout: float,
        heads: int,
        n_layers_dec: int,
        latent_size: int,
        vocab_size: int,
        bypass_bottleneck: bool,
        deconv_shape: Tuple[int, int, int],
    ) -> None:
        """Construct TransformerDecoder.

        Args:
            hidden_size: hidden size.
            ff_size: feed forward size.
            seq_len: sequence length.
            dropout: dropout rate.
            heads: number of heads.
            n_layers_enc: number of encoding layers.
            latent_size: latent size.
            vocab_size: vocabulary size.
            bypass_bottleneck: whether the bottleneck should be by passed.
            deconv_shape: shape of the deconvoluted samples. A tuple with three
                dimensions.
        """
        super().__init__()

        self.position = PositionalEncoding(hidden_size, dropout)
        self.embedding = nn.Sequential(
            Embeddings(hidden_size, vocab_size), self.position
        )
        self.attn_enc = MultiHeadedAttention(heads, hidden_size)
        self.ff_enc = PositionwiseFeedForward(hidden_size, ff_size, dropout)
        self.attn_dec_1 = MultiHeadedAttention(heads, hidden_size)
        self.attn_dec_2 = MultiHeadedAttention(heads, hidden_size)

        self.ff_dec = PositionwiseFeedForward(hidden_size, ff_size, dropout)

        encoder_layers = TransformerEncoderLayer(
            hidden_size, seq_len, self.attn_enc, self.ff_enc, dropout
        )
        decoder_layers = TransformerDecoderLayer(
            hidden_size,
            seq_len,
            self.attn_dec_1,
            self.attn_dec_2,
            self.ff_dec,
            dropout,
        )

        self.final_encodes = clones(encoder_layers, 1)
        self.layers = clones(decoder_layers, n_layers_dec)
        self.norm = LayerNorm(hidden_size)
        self.bypass_bottleneck = bypass_bottleneck
        self.hidden_size = hidden_size
        self.seq_len = seq_len
        self.outputs2vocab = torch.nn.Linear(hidden_size, vocab_size)
        self.deconv_shape = deconv_shape
        self.deconv_bottleneck = DeconvBottleneck(
            hidden_size, seq_len=seq_len, dim_factor=deconv_shape[2]
        )
        self.linear = nn.Linear(latent_size, deconv_shape[2] * deconv_shape[1])

    def forward(
        self,
        x: torch.Tensor,
        mem: torch.Tensor,
        src_mask: torch.Tensor,
        tgt_mask: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass in the model.

        Args:
            x: input tensor.
            mem: memory tensor.
            src_mask: source sequence mask.
            tgt_mask: target sequence mask.

        Returns:
            model output.
        """
        x = self.embedding(x)
        if not self.bypass_bottleneck:
            mem = F.relu(self.linear(mem))
            mem = mem.view(-1, 64, self.deconv_shape[2])
            mem = self.deconv_bottleneck(mem)
            mem = mem.permute(0, 2, 1)
        for final_encode in self.final_encodes:  # type:ignore
            mem = final_encode(mem, src_mask)
        mem = self.norm(mem)
        for _, attn_layer in enumerate(self.layers):  # type:ignore
            x = attn_layer(x, mem, mem, src_mask, tgt_mask)
        x = self.norm(x)
        x = self.outputs2vocab(F.relu(x))
        return x

    def inference_direct(
        self,
        latent: torch.Tensor,
        mask_lengths: torch.Tensor,
        tokenizer: Tokenizer,
    ) -> Tuple[List[str], torch.Tensor]:
        """Direct inference from latent space.

        Args:
            latent: latent tensor.
            mask_lengths: masking tensor.
            tokenizer: tokenizer.

        Returns:
            a tuple containing decoded strings and indices.
        """
        device = get_device_from_tensor(latent)
        batch_size = latent.size(0)
        token_indices = torch.full(
            (batch_size, 1), tokenizer.sos_token_id, device=device
        ).long()

        src_mask = torch.zeros((latent.shape[0], 1, self.seq_len), device=device)

        for index in range(mask_lengths.shape[0]):
            mask_len = int(mask_lengths[index].item())
            src_mask[index, :, :mask_len] = torch.ones((1, 1, mask_len), device=device)
        self.eval()
        for i in range(self.seq_len - 1):
            trg_mask = subsequent_mask(token_indices.size(1)).long().to(device)
            logits = self(
                torch.autograd.Variable(token_indices), latent, src_mask, trg_mask
            )

            prob = F.softmax(logits[:, i, :], dim=-1)
            _, next_token = torch.max(prob, dim=1)

            next_token = next_token.unsqueeze(1)
            token_indices = torch.cat([token_indices, next_token], dim=1)

        decoded_texts = []
        for index in range(batch_size):
            tokens = [
                tokenizer.convert_id_to_token(vocab_index.item())
                for vocab_index in token_indices[index]
            ]
            text = "".join(tokens).split()[0]
            decoded_texts.append(text)
        return decoded_texts, token_indices


class TransformerDecoderLayer(nn.Module):
    """Self-attention/source-attention/feedforward implementation."""

    def __init__(
        self,
        size: int,
        seq_len: int,
        self_attn: nn.Module,
        src_attn: nn.Module,
        feed_forward: nn.Module,
        dropout: float,
    ) -> None:
        """Construct TransformerDecoderLayer.

        Args:
            size: model size.
            seq_len: sequence length.
            self_attn: self-attention layer.
            src_attn: source attention layer.
            feed_forward: feed forward layer.
            dropout: droupout rate.
        """
        super().__init__()
        self.size = size
        self.tgt_len = seq_len
        self.self_attn = self_attn
        self.src_attn = src_attn
        self.feed_forward = feed_forward
        self.sublayer = clones(SublayerConnection(self.size, dropout), 3)

    def forward(
        self,
        x: torch.Tensor,
        memory_key: torch.Tensor,
        memory_val: torch.Tensor,
        src_mask: torch.Tensor,
        tgt_mask: torch.Tensor,
        return_attn: bool = False,
    ) -> Any:
        """Forward pass in the model.

        Args:
            x: input tensor
            memory_key: memory key tensor.
            memory_val: memory value tensor.s
            src_mask: mask to apply in the source attention layer.
            tgt_mask: mask to apply in the target attention layer.
            return_attn: whether to return the attention together with the output.
                Defaults to False, return only encoder output.

        Returns:
            model output.
        """
        m_key = memory_key
        m_val = memory_val
        if return_attn:
            x = self.sublayer[0](  # type:ignore
                x, lambda x: self.self_attn(x, x, x, tgt_mask)
            )
            src_attn = self.src_attn(x, m_key, m_val, src_mask, return_attn=True)
            x = self.sublayer[1](  # type:ignore
                x, lambda x: self.src_attn(x, m_key, m_val, src_mask)
            )
            return self.sublayer[2](x, self.feed_forward), src_attn  # type:ignore
        else:
            x = self.sublayer[0](  # type:ignore
                x, lambda x: self.self_attn(x, x, x, tgt_mask)
            )
            x = self.sublayer[1](  # type:ignore
                x, lambda x: self.src_attn(x, m_key, m_val, src_mask)
            )
            return self.sublayer[2](x, self.feed_forward)  # type:ignore



"""Moses AAE implementation."""

import argparse

from guacamol_baselines.moses_baselines.aae_distribution_learning import AaeGenerator


class AAE:
    def __init__(
        self,
        model_path: str,
        model_config_path: str,
        vocab_path: str,
        n_samples: int,
        n_batch: int,
        max_len: int,
        device: str = "cpu",
    ):
        """Initialize AAE.

        Args:
            model_path: path from where to load the model.
            model_config_path: path from where to load the model config.
            vocab_path: path from where to load the vocab.
            n_samples: number of samples to sample.
            n_batch: size of the batch.
            max_len: max length of SMILES.
            device: device used for computation. Defaults to cpu.
        """
        self.config = argparse.Namespace(
            model_load=model_path,
            config_load=model_config_path,
            vocab_load=vocab_path,
            n_samples=n_samples,
            n_batch=n_batch,
            max_len=max_len,
            device=device,
        )

    def get_generator(self) -> AaeGenerator:
        """Create an instance of the AaeGenerator.

        Returns:
            an instance of AaeGenerator.
        """
        optimiser = AaeGenerator(self.config)
        return optimiser



"""TorchDrugGenerationAlgorithm initialization."""
from .checks import TORCH_OPENMP_CHECK, torch
from .core import TorchDrugGCPN, TorchDrugGenerator, TorchDrugGraphAF

# re-enable openMP after torchdrug imports if applicable
if TORCH_OPENMP_CHECK:
    torch._C.has_openmp = True

__all__ = ["TorchDrugGenerator", "TorchDrugGCPN", "TorchDrugGraphAF"]



"""S3 storage tests.

These assume a bucket `gt4sd-ci-tests` and an artifact `a_folder/containing/a_file.txt`.
"""

import logging
import os
import shutil
import tempfile

import pytest

from gt4sd.s3 import GT4SDS3Client, sync_folder_with_s3
from gt4sd.tests.utils import GT4SDTestSettings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

test_settings = GT4SDTestSettings.get_instance()


@pytest.fixture
def client() -> GT4SDS3Client:
    return GT4SDS3Client(
        host=test_settings.gt4sd_s3_host,
        access_key=test_settings.gt4sd_s3_access_key,
        secret_key=test_settings.gt4sd_s3_secret_key,
        secure=test_settings.gt4sd_s3_secure,
    )


def test_bucket_listing(client: GT4SDS3Client):
    bucket_names = client.list_bucket_names()
    assert isinstance(bucket_names, list)
    assert "gt4sd-ci-tests" in bucket_names


def test_object_listing(client: GT4SDS3Client):
    object_names = client.list_object_names(bucket="gt4sd-ci-tests", prefix="a_folder")
    assert isinstance(object_names, list)
    for object_name in object_names:
        assert object_name.endswith(".txt")


def test_directory_listing(client: GT4SDS3Client):
    object_names = client.list_directories(bucket="gt4sd-ci-tests", prefix="a_folder")
    assert isinstance(object_names, set)
    for object_name in object_names:
        assert object_name == "containing"


def test_sync_folder(client: GT4SDS3Client):
    directory = tempfile.mkdtemp()
    try:
        client.sync_folder("gt4sd-ci-tests", directory)
        filepaths = set(
            [filepath.replace(directory, "") for filepath in os.listdir(directory)]
        )
        objectpaths = set(
            map(
                lambda path: path.split("/")[0],
                client.list_object_names(bucket="gt4sd-ci-tests"),
            )
        )
        assert filepaths == objectpaths
    finally:
        logger.info(f"cleaning up test folder {directory}")
        shutil.rmtree(directory)


def test_sync_folder_with_s3(client):
    directory = tempfile.mkdtemp()
    try:
        sync_folder_with_s3(
            host=test_settings.gt4sd_s3_host,
            access_key=test_settings.gt4sd_s3_access_key,
            secret_key=test_settings.gt4sd_s3_secret_key,
            bucket="gt4sd-ci-tests",
            folder_path=directory,
            secure=test_settings.gt4sd_s3_secure,
        )
        filepaths = set(
            [filepath.replace(directory, "") for filepath in os.listdir(directory)]
        )
        objectpaths = set(
            map(
                lambda path: path.split("/")[0],
                client.list_object_names(bucket="gt4sd-ci-tests"),
            )
        )
        assert filepaths == objectpaths
    finally:
        logger.info(f"cleaning up test folder {directory}")
        shutil.rmtree(directory)



"""Controlled Latent attribute Space Sampling initialization."""
import logging

from ....extras import EXTRAS_ENABLED

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if EXTRAS_ENABLED:
    from .core import PAG, CLaSS, CogMol

    __all__ = ["CLaSS", "CogMol", "PAG"]
else:
    logger.warning("install cogmol-inference extras to use CLaSS")
